import numpy as np
import itertools


class Structure:

    def __init__(self,
                 positions=None,
                 scaled_positions=None,
                 masses=None,
                 charges=None,
                 cell=None,
                 force_sets=None,
                 force_constants=None,
                 atomic_numbers=None,
                 atomic_elements=None,
                 atom_type_index=None,
                 primitive_matrix=None):

        """
        :param positions: atoms cartesian positions (array Ndim x Natoms)
        :param scaled_positions: atom positions scaled to 1 (array Ndim x Natoms)
        :param masses: masses of the atoms (vector NAtoms)
        :param cell: Numpy array containing the unit cell (lattice vectors in rows)
        :param force_sets: force constants: Harmonic force constants
        :param force_constants: atomic numbers vector (1x Natoms):
        :param atomic_numbers: number of total atoms in the crystal:
        :param atomic_elements: atomic names of each element (ex: H, Be, Si,..) (vector Natoms):
        :param atom_type_index: index vector that contains the number of different types of atoms in crystal (vector NdiferentAtoms)
        :param primitive_matrix: matrix that defines the primitive cell respect to the unicell
        """

        self._cell = np.array(cell, dtype='double')
        self._masses = np.array(masses, dtype='double')
        self._atomic_numbers = np.array(atomic_numbers, dtype='double')

        self._force_constants = force_constants
        self._force_sets = force_sets
        self._atomic_elements = atomic_elements
        self._atom_type_index = atom_type_index
        self._scaled_positions = scaled_positions
        self._positions = positions
        self._primitive_matrix = primitive_matrix
        self._charges = charges  # Only for LAMMPS supercell generation

        self._primitive_cell = None
        self._supercell_matrix = None
        self._supercell_phonon = None
        self._supercell_phonon_renormalized = None
        self._number_of_cell_atoms = None
        self._number_of_atoms = None
        self._number_of_atom_types = None
        self._number_of_primitive_atoms = None

        # Get atomic types from masses if available
        if atomic_elements is None and masses is not None:
            self._atomic_elements = []
            for i in masses:
                for j in atom_data:
                    if "{0:.1f}".format(i) == "{0:.1f}".format(j[3]):
                        self._atomic_elements.append(j[1])

        # Get atomic numbers and masses from available information
        if atomic_numbers is None and self._atomic_elements is not None:
            self._atomic_numbers = np.array([symbol_map[i] for i in self._atomic_elements])
        else:
            self._atomic_numbers = np.array(atomic_numbers)

        if masses is None and self._atomic_numbers is not None:
            self._masses = np.array([atom_data[i][3] for i in self._atomic_numbers])
        else:
            self._masses = masses

    # -------- Methods start here -----------

    # Getting data
    def get_data_from_dict(self, data_dictionary):
        for data in self.__dict__:
            try:
                self.__dict__[data] = data_dictionary[data]
            except KeyError:
                continue

    # Cell related methods
    def set_cell(self, cell):
        self._cell = np.array(cell, dtype='double')

    def get_cell(self, supercell=None):

        if supercell is not None:
            return np.dot(self._cell, np.diag(supercell))

        return self._cell

    def get_supercell_phonon(self):

        if self.get_force_constants() is not None:
            supercell_phonon = self.get_force_constants().get_supercell()
        elif self.get_force_sets() is not None:
            supercell_phonon = self.get_force_sets().get_supercell()
        else:
            supercell_phonon = np.identity(self.get_number_of_dimensions(), dtype=int)
        return supercell_phonon

    def set_supercell_matrix(self, supercell_matrix):
        self._supercell_matrix = supercell_matrix

    def get_supercell_matrix(self):
        if self._supercell_matrix is None:
            self._supercell_matrix = np.array(self.get_number_of_dimensions() * [1], dtype=int)
        return self._supercell_matrix

    def get_primitive_cell(self):
        if self._primitive_cell is None:
            self._primitive_cell = np.dot(self.get_cell().T, self.get_primitive_matrix()).T
        return self._primitive_cell

    # Cell matrix related methods
    def set_primitive_matrix(self,primitive_matrix):
        self._primitive_matrix = primitive_matrix
        self._number_of_atom_types = None
        self._number_of_primitive_atoms = None
        self._atom_type_index = None
        self._primitive_cell = None

    def get_primitive_matrix(self):
        if self._primitive_matrix is None:
            if self._primitive_cell is None:
                print('Warning: No primitive matrix defined! Using unit cell as primitive')
                self._primitive_matrix = np.identity(self.get_number_of_dimensions())
            else:
                self._primitive_matrix = np.dot(np.linalg.inv(self.get_cell()), self._primitive_cell)
        return self._primitive_matrix

    # Positions related methods
    def set_positions(self, cart_positions):
        self._scaled_positions = np.dot(cart_positions, np.linalg.inv(self.get_cell()))

    def get_positions(self, supercell=None):
        if self._positions is None:
            if self._scaled_positions is None:
                print('No positions provided')
                exit()
            else:
                self._positions = np.dot(self._scaled_positions, self.get_cell())

        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        position_supercell = []
        for k in range(self._positions.shape[0]):
            for r in itertools.product(*[range (i) for i in supercell[::-1]]):
                position_supercell.append(self._positions[k,:] + np.dot(np.array(r[::-1]), self.get_cell()))
        position_supercell = np.array(position_supercell)

        return position_supercell

    def get_scaled_positions(self, supercell=None):

        if self._scaled_positions is None:
            self._scaled_positions = np.dot(self.get_positions(), np.linalg.inv(self.get_cell()))

        if supercell is not None:
            cell = self.get_cell(supercell=supercell)
            scaled_positions = np.dot(self.get_positions(supercell), np.linalg.inv(cell))
            return scaled_positions

        return self._scaled_positions

    # Force related methods
    def forces_available(self):
        if self.get_force_constants() is not None or self.get_force_sets() is not None:
            return True
        else:
            return False

    def set_force_constants(self, force_constants):
        self._force_constants = force_constants

    def get_force_constants(self):
        return self._force_constants

    def set_force_set(self, force_set):
        self._force_sets = force_set

    def get_force_sets(self):

        if not isinstance(self._force_sets,type(None)):
            force_atoms_file = self._force_sets.get_dict()['natom']
            force_atoms_input = np.prod(np.diagonal(self._force_sets.get_supercell())) * self.get_number_of_atoms()

            if force_atoms_file != force_atoms_input:
                print("Error: FORCE_SETS file does not match with SUPERCELL MATRIX")
                exit()

        return self._force_sets

    def set_masses(self, masses):
        self._masses = np.array(masses, dtype='double')

    def get_masses(self, supercell=None):
        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        mass_supercell = []
        for j in range(self.get_number_of_cell_atoms()):
            mass_supercell += [ self._masses[j] ] * np.prod(supercell)
        return mass_supercell

    def get_charges(self, supercell=None):

        if self._charges is None:
            return None

        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        charges_supercell = []
        for j in range(self.get_number_of_cell_atoms()):
            charges_supercell += [ self._charges[j] ] * np.prod(supercell)
        return charges_supercell

    # Number of atoms and dimensions related methods
    def get_number_of_cell_atoms(self):
        if self._number_of_cell_atoms is None:
            self._number_of_cell_atoms = self.get_positions().shape[0]
        return self._number_of_cell_atoms

    def get_number_of_dimensions(self):
        return self.get_cell().shape[1]

    def get_atomic_numbers(self, supercell=None):
        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        atomic_numbers_supercell = []
        for j in range(self.get_number_of_cell_atoms()):
            atomic_numbers_supercell += [self._atomic_numbers[j] ] * np.prod(supercell)

        return np.array(atomic_numbers_supercell,dtype=int)

    def get_number_of_atoms(self):
        if self._number_of_atoms is None:
            self._number_of_atoms = self.get_number_of_cell_atoms()
        return self._number_of_atoms

    def set_number_of_atoms(self,number_of_atoms):
        self._number_of_atoms = number_of_atoms

    def get_number_of_atom_types(self):
        if self._number_of_atom_types is None:
            self._number_of_atom_types = len(set(self.get_atom_type_index()))
    #         print(self._number_of_atom_types)
        return self._number_of_atom_types

    def get_number_of_primitive_atoms(self):
        if self._number_of_primitive_atoms is None:
            self._number_of_primitive_atoms = len(set(self.get_atom_type_index()))
        return self._number_of_primitive_atoms

    def set_number_of_primitive_atoms(self,number_of_primitive_atoms):
        self._number_of_primitive_atoms = number_of_primitive_atoms

    # Atomic types related methods
    def get_atomic_elements(self, supercell=None, unique=False):
        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        atomic_types = []
        for j in range(self.get_number_of_cell_atoms()):
            atomic_types += [self._atomic_elements[j]] * np.prod(supercell)
        if unique:
            unique_indices = np.unique(self.get_atom_type_index(supercell=supercell), return_index=True)[1]
            atomic_types =  np.array(atomic_types)[unique_indices]
        return atomic_types

    def set_atom_type_index_by_element(self):
        if self._atom_type_index is None:
            self._atom_type_index = self._atomic_numbers.copy()
            for i in range(self.get_number_of_cell_atoms()):
                for index in range(self.get_number_of_atom_types()):
                    if self._atomic_numbers[index] == self._atomic_numbers[i]:
                        self._atom_type_index[i] = index

    def get_atom_type_index(self, supercell=None):

        # Tolerance for accepting equivalent atoms in super cell
        masses = self.get_masses(supercell=supercell)
        tolerance = 0.001
        if self._atom_type_index is None:
            primitive_cell_inverse = np.linalg.inv(self.get_primitive_cell())

            self._atom_type_index = np.array(self.get_number_of_cell_atoms() * [None])
            counter = 0
            for i in range(self.get_number_of_cell_atoms()):
                if self._atom_type_index[i] is None:
                    self._atom_type_index[i] = counter
                    counter += 1
                for j in range(i+1, self.get_number_of_cell_atoms()):
                    coordinates_atom_i = self.get_positions()[i]
                    coordinates_atom_j = self.get_positions()[j]

                    difference_in_cell_coordinates = np.around((np.dot(primitive_cell_inverse.T, (coordinates_atom_j - coordinates_atom_i))))
                    projected_coordinates_atom_j = coordinates_atom_j - np.dot(self.get_primitive_cell().T, difference_in_cell_coordinates)
                    separation = pow(np.linalg.norm(projected_coordinates_atom_j - coordinates_atom_i),2)

                    if separation < tolerance and masses[i] == masses[j]:
                        self._atom_type_index[j] = self._atom_type_index[i]
        self._atom_type_index = np.array(self._atom_type_index,dtype=int)

        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        atom_type_index_supercell = []
        for j in range(self.get_number_of_cell_atoms()):
            atom_type_index_supercell += [self._atom_type_index[j] ] * np.prod(supercell)
        return atom_type_index_supercell

    def get_cell_parameters(self, supercell=None):

        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        cell = self.get_cell(supercell=supercell)

        a = np.linalg.norm(cell[0])
        b = np.linalg.norm(cell[1])
        c = np.linalg.norm(cell[2])

        alpha = np.arccos(np.dot(cell[1], cell[2])/(c*b))
        gamma = np.arccos(np.dot(cell[1], cell[0])/(a*b))
        beta = np.arccos(np.dot(cell[2], cell[0])/(a*c))

        return a, b, c, alpha, beta, gamma

    def get_commensurate_points(self, supercell=None):

        if supercell is None:
            supercell = self.get_number_of_dimensions() * [1]

        primitive_matrix = self.get_primitive_matrix()

        commensurate_points = []
        for k1 in np.arange(-0.5, 0.5, 1./(supercell[0]*2)):
            for k2 in np.arange(-0.5, 0.5, 1./(supercell[1]*2)):
                for k3 in np.arange(-0.5, 0.5, 1./(supercell[2]*2)):
    
                    q_point = [np.around(k1,decimals=5), np.around(k2,decimals=5), np.around(k3,decimals=5)]
    
                    q_point_unit_cell = np.dot(q_point, np.linalg.inv(primitive_matrix))
                    q_point_unit_cell = np.multiply(q_point_unit_cell, supercell) * 2
    
                    if np.all(np.equal(np.mod(q_point_unit_cell, 1), 0)):
                        commensurate_points.append(q_point)

        return commensurate_points

    def get_path_using_seek_path(self):
        try:
            import seekpath

            cell = self.get_cell()
            positions = self.get_scaled_positions()
            numbers = np.unique(self.get_atomic_elements(), return_inverse=True)[1]
            structure = (cell, positions, numbers)
            path_data = seekpath.get_path(structure)

            labels = path_data['point_coords']

            band_ranges = []
            for set in path_data['path']:
                band_ranges.append([labels[set[0]], labels[set[1]]])

            return {'ranges': band_ranges,
                    'labels': path_data['path']}
        except ImportError:
            print ('Seekpath not installed. Autopath is deactivated')
            band_ranges=([[[0.0, 0.0, 0.0], [0.5, 0.0, 0.5]]])
            return {'ranges': band_ranges,
                    'labels': [['GAMMA', '1/2,0,1/2']]}


atom_data = [
    [  0, "X", "X", 0],
    [  1, "H", "Hydrogen", 1.00794],
    [  2, "He", "Helium", 4.002602],
    [  3, "Li", "Lithium", 6.941],
    [  4, "Be", "Beryllium", 9.012182],
    [  5, "B", "Boron", 10.811],
    [  6, "C", "Carbon", 12.0107],
    [  7, "N", "Nitrogen", 14.0067],
    [  8, "O", "Oxygen", 15.9994],
    [  9, "F", "Fluorine", 18.9984032],
    [ 10, "Ne", "Neon", 20.1797],
    [ 11, "Na", "Sodium", 22.98976928],
    [ 12, "Mg", "Magnesium", 24.3050],
    [ 13, "Al", "Aluminium", 26.9815386],
    [ 14, "Si", "Silicon", 28.0855],
    [ 15, "P", "Phosphorus", 30.973762],
    [ 16, "S", "Sulfur", 32.065],
    [ 17, "Cl", "Chlorine", 35.453],
    [ 18, "Ar", "Argon", 39.948],
    [ 19, "K", "Potassium", 39.0983],
    [ 20, "Ca", "Calcium", 40.078],
    [ 21, "Sc", "Scandium", 44.955912],
    [ 22, "Ti", "Titanium", 47.867],
    [ 23, "V", "Vanadium", 50.9415],
    [ 24, "Cr", "Chromium", 51.9961],
    [ 25, "Mn", "Manganese", 54.938045],
    [ 26, "Fe", "Iron", 55.845],
    [ 27, "Co", "Cobalt", 58.933195],
    [ 28, "Ni", "Nickel", 58.6934],
    [ 29, "Cu", "Copper", 63.546],
    [ 30, "Zn", "Zinc", 65.38],
    [ 31, "Ga", "Gallium", 69.723],
    [ 32, "Ge", "Germanium", 72.64],
    [ 33, "As", "Arsenic", 74.92160],
    [ 34, "Se", "Selenium", 78.96],
    [ 35, "Br", "Bromine", 79.904],
    [ 36, "Kr", "Krypton", 83.798],
    [ 37, "Rb", "Rubidium", 85.4678],
    [ 38, "Sr", "Strontium", 87.62],
    [ 39, "Y", "Yttrium", 88.90585],
    [ 40, "Zr", "Zirconium", 91.224],
    [ 41, "Nb", "Niobium", 92.90638],
    [ 42, "Mo", "Molybdenum", 95.96],
    [ 43, "Tc", "Technetium", 0],
    [ 44, "Ru", "Ruthenium", 101.07],
    [ 45, "Rh", "Rhodium", 102.90550],
    [ 46, "Pd", "Palladium", 106.42],
    [ 47, "Ag", "Silver", 107.8682],
    [ 48, "Cd", "Cadmium", 112.411],
    [ 49, "In", "Indium", 114.818],
    [ 50, "Sn", "Tin", 118.710],
    [ 51, "Sb", "Antimony", 121.760],
    [ 52, "Te", "Tellurium", 127.60],
    [ 53, "I", "Iodine", 126.90447],
    [ 54, "Xe", "Xenon", 131.293],
    [ 55, "Cs", "Caesium", 132.9054519],
    [ 56, "Ba", "Barium", 137.327],
    [ 57, "La", "Lanthanum", 138.90547],
    [ 58, "Ce", "Cerium", 140.116],
    [ 59, "Pr", "Praseodymium", 140.90765],
    [ 60, "Nd", "Neodymium", 144.242],
    [ 61, "Pm", "Promethium", 0],
    [ 62, "Sm", "Samarium", 150.36],
    [ 63, "Eu", "Europium", 151.964],
    [ 64, "Gd", "Gadolinium", 157.25],
    [ 65, "Tb", "Terbium", 158.92535],
    [ 66, "Dy", "Dysprosium", 162.500],
    [ 67, "Ho", "Holmium", 164.93032],
    [ 68, "Er", "Erbium", 167.259],
    [ 69, "Tm", "Thulium", 168.93421],
    [ 70, "Yb", "Ytterbium", 173.054],
    [ 71, "Lu", "Lutetium", 174.9668],
    [ 72, "Hf", "Hafnium", 178.49],
    [ 73, "Ta", "Tantalum", 180.94788],
    [ 74, "W", "Tungsten", 183.84],
    [ 75, "Re", "Rhenium", 186.207],
    [ 76, "Os", "Osmium", 190.23],
    [ 77, "Ir", "Iridium", 192.217],
    [ 78, "Pt", "Platinum", 195.084],
    [ 79, "Au", "Gold", 196.966569],
    [ 80, "Hg", "Mercury", 200.59],
    [ 81, "Tl", "Thallium", 204.3833],
    [ 82, "Pb", "Lead", 207.2],
    [ 83, "Bi", "Bismuth", 208.98040],
    [ 84, "Po", "Polonium", 0],
    [ 85, "At", "Astatine", 0],
    [ 86, "Rn", "Radon", 0],
    [ 87, "Fr", "Francium", 0],
    [ 88, "Ra", "Radium", 0],
    [ 89, "Ac", "Actinium", 0],
    [ 90, "Th", "Thorium", 232.03806],
    [ 91, "Pa", "Protactinium", 231.03588],
    [ 92, "U", "Uranium", 238.02891],
    [ 93, "Np", "Neptunium", 0],
    [ 94, "Pu", "Plutonium", 0],
    [ 95, "Am", "Americium", 0],
    [ 96, "Cm", "Curium", 0],
    [ 97, "Bk", "Berkelium", 0],
    [ 98, "Cf", "Californium", 0],
    [ 99, "Es", "Einsteinium", 0],
    [100, "Fm", "Fermium", 0],
    [101, "Md", "Mendelevium", 0],
    [102, "No", "Nobelium", 0],
    [103, "Lr", "Lawrencium", 0],
    [104, "Rf", "Rutherfordium", 0],
    [105, "Db", "Dubnium", 0],
    [106, "Sg", "Seaborgium", 0],
    [107, "Bh", "Bohrium", 0],
    [108, "Hs", "Hassium", 0],
    [109, "Mt", "Meitnerium", 0],
    [110, "Ds", "Darmstadtium", 0],
    [111, "Rg", "Roentgenium", 0],
    [112, "Cn", "Copernicium", 0],
    [113, "Uut", "Ununtrium", 0],
    [114, "Uuq", "Ununquadium", 0],
    [115, "Uup", "Ununpentium", 0],
    [116, "Uuh", "Ununhexium", 0],
    [117, "Uus", "Ununseptium", 0],
    [118, "Uuo", "Ununoctium", 0],
    ]

symbol_map = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Uut": 113,
    "Uuq": 114,
    "Uup": 115,
    "Uuh": 116,
    "Uus": 117,
    "Uuo": 118,
    }



"""Matrix utility functions for setting rotational invariants."""

# from typing import Optional
#
import itertools

import numpy as np
import scipy
from pypolymlp.core.interface_vasp import Poscar
from pypolymlp.utils.structure_utils import supercell_diagonal
from scipy.sparse import csr_array

from macer.externals.symfc_bundled import Symfc
from macer.externals.symfc_bundled.basis_sets.basis_sets_O2 import FCBasisSetO2
from macer.externals.symfc_bundled.spg_reps import SpgRepsBase
from macer.externals.symfc_bundled.utils.eig_tools import eigsh_projector, eigsh_projector_sumrule
from macer.externals.symfc_bundled.utils.utils import SymfcAtoms
from macer.externals.symfc_bundled.utils.utils_O2 import (
    get_lat_trans_compr_matrix,
    get_lat_trans_decompr_indices,
)

try:
    from macer.externals.symfc_bundled.utils.eig_tools import dot_product_sparse
except ImportError:
    pass

import copy


def _orthogonalize_constraints(positions_cartesian: np.ndarray):
    """Orthogonalize constraints derived from rotational invariance."""
    natom = positions_cartesian.shape[0]
    N3 = 3 * natom

    C = np.zeros((N3, 3), dtype="double")
    C[1::3, 0] = -positions_cartesian[:, 0]
    C[2::3, 2] = positions_cartesian[:, 0]
    C[2::3, 1] = -positions_cartesian[:, 1]
    C[0::3, 0] = positions_cartesian[:, 1]
    C[0::3, 2] = -positions_cartesian[:, 2]
    C[1::3, 1] = positions_cartesian[:, 2]

    #    # U, S, V = np.linalg.svd(C, full_matrices=False)
    #    # nonzero = np.abs(S) > 1e-15
    #    # C = U[:,nonzero]
    #    # C[np.abs(C) < 1e-15] = 0.0
    #    # print(S)
    proj_C = C @ np.linalg.inv(C.T @ C) @ C.T
    eigvals, eigvecs = np.linalg.eigh(proj_C)
    nonzero = np.isclose(eigvals, 1.0)
    C = eigvecs[:, nonzero]
    return C


def complement_sum_rule(natom, decompr_idx=None, n_lp=None):
    """Construct sum rule."""
    N33 = natom * 9
    NN33 = natom**2 * 9
    row = []
    for j, alpha, beta in itertools.product(range(natom), range(3), range(3)):
        i = np.arange(natom) * N33
        const_add = j * 9 + alpha * 3 + beta
        row.extend(i + const_add)
    col = np.repeat(np.arange(N33), natom)
    data = np.ones(len(col)) / np.sqrt(natom)

    if decompr_idx is not None:
        data /= np.sqrt(n_lp)
        c_sum = csr_array(
            (data, (decompr_idx[row], col)),
            shape=(NN33 // n_lp, N33),
            dtype="double",
        )
    else:
        c_sum = csr_array(
            (data, (row, col)),
            shape=(NN33, N33),
            dtype="double",
        )

    return c_sum


def complementary_compr_projector_rot_sum_rules_O2(
    supercell: SymfcAtoms,
    trans_perms: np.ndarray,
    n_a_compress_mat: np.ndarray,
    use_mkl: bool = False,
) -> csr_array:
    """Test function for setting rotational invariants."""
    n_lp, natom = trans_perms.shape

    # TODO: decompr_idx -> atomic_decompr_idx
    decompr_idx = get_lat_trans_decompr_indices(trans_perms)
    indep_atoms = list(range(natom))

    positions_cartesian = (supercell.scaled_positions) @ supercell.cell
    positions_cartesian[:, 0] -= np.average(positions_cartesian[:, 0])
    positions_cartesian[:, 1] -= np.average(positions_cartesian[:, 1])
    positions_cartesian[:, 2] -= np.average(positions_cartesian[:, 2])
    C2 = _orthogonalize_constraints(positions_cartesian)

    N3 = natom * 3
    N33 = N3 * 3
    NN33 = N3 * N3

    n_expand = len(indep_atoms) * 3
    data = np.tile(C2.T.reshape(-1), n_expand)
    col = np.repeat(np.arange(n_expand * 3), N3)
    row = []
    ids_ialpha = (np.arange(natom) * N33)[:, None] + (np.arange(3) * 3)[None, :]
    ids_ialpha = ids_ialpha.reshape(-1)
    for j, beta in itertools.product(indep_atoms, range(3)):
        ids = ids_ialpha + (j * 9 + beta)
        row.extend(np.tile(ids, 3))

    data /= np.sqrt(n_lp)
    c_rot_cmplt = csr_array(
        (data, (decompr_idx[row], col)),
        shape=(NN33 // n_lp, n_expand * 3),
        dtype="double",
    )

    proj = scipy.sparse.identity(NN33 // n_lp) - c_rot_cmplt @ c_rot_cmplt.T
    c_rot = eigsh_projector(proj, verbose=True)

    c_cmplt = n_a_compress_mat.T @ c_rot
    proj = c_cmplt @ c_cmplt.T
    return proj


def complement_rotational_sum_rule(supercell, indep_atoms, decompr_idx=None, n_lp=None):
    """Get complement_rotational_sum_rule."""
    positions_cartesian = (supercell.scaled_positions) @ supercell.cell
    positions_cartesian[:, 0] -= np.average(positions_cartesian[:, 0])
    positions_cartesian[:, 1] -= np.average(positions_cartesian[:, 1])
    positions_cartesian[:, 2] -= np.average(positions_cartesian[:, 2])
    C2 = _orthogonalize_constraints(positions_cartesian)

    natom = positions_cartesian.shape[0]
    N3 = natom * 3
    N33 = N3 * 3
    NN33 = N3 * N3

    n_expand = len(indep_atoms) * 3
    data = np.tile(C2.T.reshape(-1), n_expand)
    col = np.repeat(np.arange(n_expand * 3), N3)
    row = []
    ids_ialpha = (np.arange(natom) * N33)[:, None] + (np.arange(3) * 3)[None, :]
    ids_ialpha = ids_ialpha.reshape(-1)
    for j, beta in itertools.product(indep_atoms, range(3)):
        ids = ids_ialpha + (j * 9 + beta)
        row.extend(np.tile(ids, 3))

    if decompr_idx is None:
        c_rot = csr_array(
            (data, (row, col)),
            shape=(NN33, n_expand * 3),
            dtype="double",
        )
    else:
        data /= np.sqrt(n_lp)
        c_rot = csr_array(
            (data, (decompr_idx[row], col)),
            shape=(NN33 // n_lp, n_expand * 3),
            dtype="double",
        )
    return c_rot


def complementary_compr_projector_rot_O2_test(
    supercell: SymfcAtoms,
    trans_perms: np.ndarray,
    basis_set_fc2: FCBasisSetO2,
    indep_atoms=None,
    use_mkl: bool = False,
) -> csr_array:
    """Test function for setting rotational invariants."""
    n_lp, natom = trans_perms.shape
    N3 = natom * 3
    NN33 = N3 * N3

    decompr_idx = get_lat_trans_decompr_indices(trans_perms)
    c_trans = get_lat_trans_compr_matrix(decompr_idx, natom, n_lp)
    # indep_atoms = get_indep_atoms_by_lat_trans(trans_perms)
    # indep_atoms = [indep_atoms[0]]
    if indep_atoms is None:
        indep_atoms = list(range(natom))
    # indep_atoms = [2]
    # indep_atoms = [0,1,2,3,4,5,6,7]

    c_rot = complement_rotational_sum_rule(supercell, indep_atoms)
    proj_rot = c_rot @ c_rot.T

    c_sum = complement_sum_rule(natom)
    proj_sum = c_sum @ c_sum.T

    print("--------------")
    print("Only lattice translation")
    proj = dot_product_sparse(c_trans, c_trans.T, use_mkl=use_mkl)
    print(proj.shape)
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    print("--------------")
    print("Only translational sum rule")
    proj = scipy.sparse.identity(NN33) - proj_sum
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)
    c = c_trans.T @ c
    print(c.shape)
    proj = dot_product_sparse(c, c.T, use_mkl=use_mkl)
    print(proj.shape)
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    c_sum = complement_sum_rule(natom, decompr_idx=decompr_idx, n_lp=n_lp)
    proj = scipy.sparse.identity(NN33 // n_lp) - c_sum @ c_sum.T
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    #    print("--------------")
    #    print("Only rotational sum rules (indep_atoms)")
    #
    #    c_rot1 = complement_rotational_sum_rule(
    #        supercell, [0], decompr_idx=decompr_idx, n_lp=n_lp
    #    )
    #
    #    c_rot2 = complement_rotational_sum_rule(
    #        supercell, [2], decompr_idx=decompr_idx, n_lp=n_lp
    #    )
    #    #print(c_rot1 - c_rot2)

    print("--------------")
    print("Only rotational sum rules")

    c_rot = complement_rotational_sum_rule(
        supercell, indep_atoms, decompr_idx=decompr_idx, n_lp=n_lp
    )
    proj = scipy.sparse.identity(NN33) - proj_rot
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)
    c = c_trans.T @ c
    proj = dot_product_sparse(c, c.T, use_mkl=use_mkl)
    print(proj.shape)
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    proj = scipy.sparse.identity(NN33 // n_lp) - c_rot @ c_rot.T
    proj_return = copy.copy(proj)
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)
    vals, vecs = np.linalg.eigh(proj.toarray())
    print(vals)

    print("--------------")
    print("Only rotational sum rules (cmplt)")

    c = eigsh_projector(proj_rot, verbose=True)
    print(c.shape)

    print("--------------")
    print("Only translational and rotational sum rules")
    proj = scipy.sparse.identity(NN33) - proj_sum - proj_rot
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)
    c = c_trans.T @ c
    proj = dot_product_sparse(c, c.T, use_mkl=use_mkl)
    print(proj.shape)
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    proj = scipy.sparse.identity(NN33 // n_lp) - c_rot @ c_rot.T - c_sum @ c_sum.T
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    print("--------------")
    """
    Compressed projector I - P^(c)
    P^(c) = n_a_compress_mat.T @ C_trans.T @ C_sum^(c)
            @ C_sum^(c).T @ C_trans @ n_a_compress_mat
    """

    proj = scipy.sparse.identity(NN33 // n_lp) - c_rot @ c_rot.T
    c = eigsh_projector(proj, verbose=True)
    print(c.shape)

    n_a_compress_mat = np.sqrt(n_lp) * basis_set_fc2.compact_compression_matrix
    print(n_a_compress_mat.shape)
    # compress_mat = n_a_compress_mat @ basis_set_fc2.basis_set
    compress_mat = n_a_compress_mat
    print(compress_mat.shape)

    c = compress_mat.T @ c
    print(c.shape)
    proj = csr_array(c @ c.T)
    eigvecs = eigsh_projector(proj)
    print(eigvecs.shape)

    vals, vecs = np.linalg.eigh(proj.toarray())
    print(vals)
    return proj_return


if __name__ == "__main__":
    import argparse
    import signal
    import time

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    parser = argparse.ArgumentParser()
    parser.add_argument("--poscar", type=str, default=None, help="poscar")
    parser.add_argument(
        "--supercell",
        nargs=3,
        type=int,
        default=None,
        help="Supercell size (diagonal components)",
    )
    args = parser.parse_args()

    unitcell = Poscar(args.poscar).structure
    supercell = supercell_diagonal(unitcell, args.supercell)
    supercell = SymfcAtoms(
        numbers=supercell.types,
        cell=supercell.axis.T,
        scaled_positions=supercell.positions.T,
    )

    spg_reps = SpgRepsBase(supercell)
    trans_perms = spg_reps.translation_permutations
    n_lp, natom = trans_perms.shape

    symfc = Symfc(supercell, use_mkl=True, log_level=1).compute_basis_set(2)
    basis_set_fc2 = symfc.basis_set[2]
    n_a_compress_mat = np.sqrt(n_lp) * basis_set_fc2.compact_compression_matrix

    t1 = time.time()
    proj = complementary_compr_projector_rot_sum_rules_O2(
        supercell,
        trans_perms,
        n_a_compress_mat,
        use_mkl=True,
    )
    # proj = scipy.sparse.identity(proj.shape[0]) - proj
    eigvecs = eigsh_projector_sumrule(proj, verbose=True)
    t2 = time.time()
    print("Elapsed time:", t2 - t1)
    print("Number of eigenvectors:", eigvecs.shape[1])

    proj1 = complementary_compr_projector_rot_O2_test(
        supercell,
        trans_perms,
        basis_set_fc2,
        indep_atoms=[2],
    )
    proj2 = complementary_compr_projector_rot_O2_test(
        supercell,
        trans_perms,
        basis_set_fc2,
        indep_atoms=[1, 6],
    )
    # print(proj2 - proj1)



###########################################################################################
# Data parsing utilities
# Authors: Ilyes Batatia, Gregor Simm and David Kovacs
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Sequence, Tuple

import ase.data
import ase.io
import h5py
import numpy as np

from macer.externals.mace_bundled.tools import AtomicNumberTable, DefaultKeys

Positions = np.ndarray  # [..., 3]
Cell = np.ndarray  # [3,3]
Pbc = tuple  # (3,)

DEFAULT_CONFIG_TYPE = "Default"
DEFAULT_CONFIG_TYPE_WEIGHTS = {DEFAULT_CONFIG_TYPE: 1.0}


@dataclass
class KeySpecification:
    info_keys: Dict[str, str] = field(default_factory=dict)
    arrays_keys: Dict[str, str] = field(default_factory=dict)

    def update(
        self,
        info_keys: Optional[Dict[str, str]] = None,
        arrays_keys: Optional[Dict[str, str]] = None,
    ):
        if info_keys is not None:
            self.info_keys.update(info_keys)
        if arrays_keys is not None:
            self.arrays_keys.update(arrays_keys)
        return self

    @classmethod
    def from_defaults(cls):
        instance = cls()
        return update_keyspec_from_kwargs(instance, DefaultKeys.keydict())


def update_keyspec_from_kwargs(
    keyspec: KeySpecification, keydict: Dict[str, str]
) -> KeySpecification:
    # convert command line style property_key arguments into a keyspec
    infos = [
        "energy_key",
        "stress_key",
        "virials_key",
        "dipole_key",
        "head_key",
        "elec_temp_key",
        "total_charge_key",
        "polarizability_key",
        "total_spin_key",
    ]
    arrays = ["forces_key", "charges_key"]
    info_keys = {}
    arrays_keys = {}
    for key in infos:
        if key in keydict:
            info_keys[key[:-4]] = keydict[key]
    for key in arrays:
        if key in keydict:
            arrays_keys[key[:-4]] = keydict[key]
    keyspec.update(info_keys=info_keys, arrays_keys=arrays_keys)
    return keyspec


@dataclass
class Configuration:
    atomic_numbers: np.ndarray
    positions: Positions  # Angstrom
    properties: Dict[str, Any]
    property_weights: Dict[str, float]
    cell: Optional[Cell] = None
    pbc: Optional[Pbc] = None

    weight: float = 1.0  # weight of config in loss
    config_type: str = DEFAULT_CONFIG_TYPE  # config_type of config
    head: str = "Default"  # head used to compute the config


Configurations = List[Configuration]


def random_train_valid_split(
    items: Sequence, valid_fraction: float, seed: int, work_dir: str
) -> Tuple[List, List]:
    assert 0.0 < valid_fraction < 1.0

    size = len(items)
    # guarantee at least one validation, mostly for tests with tiny fitting databases
    assert size > 1
    train_size = min(size - int(valid_fraction * size), size - 1)

    indices = list(range(size))
    rng = np.random.default_rng(seed)
    rng.shuffle(indices)
    if len(indices[train_size:]) < 10:
        logging.info(
            f"Using random {100 * valid_fraction:.0f}% of training set for validation with following indices: {indices[train_size:]}"
        )
    else:
        # Save indices to file
        with open(work_dir + f"/valid_indices_{seed}.txt", "w", encoding="utf-8") as f:
            for index in indices[train_size:]:
                f.write(f"{index}\n")

        logging.info(
            f"Using random {100 * valid_fraction:.0f}% of training set for validation with indices saved in: {work_dir}/valid_indices_{seed}.txt"
        )

    return (
        [items[i] for i in indices[:train_size]],
        [items[i] for i in indices[train_size:]],
    )


def config_from_atoms_list(
    atoms_list: List[ase.Atoms],
    key_specification: KeySpecification,
    config_type_weights: Optional[Dict[str, float]] = None,
    head_name: str = "Default",
) -> Configurations:
    """Convert list of ase.Atoms into Configurations"""
    if config_type_weights is None:
        config_type_weights = DEFAULT_CONFIG_TYPE_WEIGHTS

    all_configs = []
    for atoms in atoms_list:
        all_configs.append(
            config_from_atoms(
                atoms,
                key_specification=key_specification,
                config_type_weights=config_type_weights,
                head_name=head_name,
            )
        )
    return all_configs


def config_from_atoms(
    atoms: ase.Atoms,
    key_specification: KeySpecification = KeySpecification(),
    config_type_weights: Optional[Dict[str, float]] = None,
    head_name: str = "Default",
) -> Configuration:
    """Convert ase.Atoms to Configuration"""
    if config_type_weights is None:
        config_type_weights = DEFAULT_CONFIG_TYPE_WEIGHTS

    atomic_numbers = np.array(
        [ase.data.atomic_numbers[symbol] for symbol in atoms.symbols]
    )
    pbc = tuple(atoms.get_pbc())
    cell = np.array(atoms.get_cell())
    config_type = atoms.info.get("config_type", "Default")
    weight = atoms.info.get("config_weight", 1.0) * config_type_weights.get(
        config_type, 1.0
    )

    properties = {}
    property_weights = {}
    for name in list(key_specification.arrays_keys) + list(key_specification.info_keys):
        property_weights[name] = atoms.info.get(f"config_{name}_weight", 1.0)

    for name, atoms_key in key_specification.info_keys.items():
        properties[name] = atoms.info.get(atoms_key, None)
        if not atoms_key in atoms.info:
            property_weights[name] = 0.0

    for name, atoms_key in key_specification.arrays_keys.items():
        properties[name] = atoms.arrays.get(atoms_key, None)
        if not atoms_key in atoms.arrays:
            property_weights[name] = 0.0

    return Configuration(
        atomic_numbers=atomic_numbers,
        positions=atoms.get_positions(),
        properties=properties,
        weight=weight,
        property_weights=property_weights,
        head=head_name,
        config_type=config_type,
        pbc=pbc,
        cell=cell,
    )


def test_config_types(
    test_configs: Configurations,
) -> List[Tuple[str, List[Configuration]]]:
    """Split test set based on config_type-s"""
    test_by_ct = []
    all_cts = []
    for conf in test_configs:
        if conf.head is None:
            conf.head = ""
        config_type_name = conf.config_type + "_" + conf.head
        if config_type_name not in all_cts:
            all_cts.append(config_type_name)
            test_by_ct.append((config_type_name, [conf]))
        else:
            ind = all_cts.index(config_type_name)
            test_by_ct[ind][1].append(conf)
    return test_by_ct


def load_from_xyz(
    file_path: str,
    key_specification: KeySpecification,
    head_name: str = "Default",
    config_type_weights: Optional[Dict] = None,
    extract_atomic_energies: bool = False,
    keep_isolated_atoms: bool = False,
    no_data_ok: bool = False,
) -> Tuple[Dict[int, float], Configurations]:
    atoms_list = ase.io.read(file_path, index=":")
    energy_key = key_specification.info_keys["energy"]
    forces_key = key_specification.arrays_keys["forces"]
    stress_key = key_specification.info_keys["stress"]
    head_key = key_specification.info_keys["head"]
    original_energy_key = energy_key
    original_forces_key = forces_key
    original_stress_key = stress_key
    if energy_key == "energy":
        logging.warning(
            "Since ASE version 3.23.0b1, using energy_key 'energy' is no longer safe when communicating between MACE and ASE. We recommend using a different key, rewriting 'energy' to 'REF_energy'. You need to use --energy_key='REF_energy' to specify the chosen key name."
        )
        key_specification.info_keys["energy"] = "REF_energy"
        for atoms in atoms_list:
            try:
                # print("OK")
                atoms.info["REF_energy"] = atoms.get_potential_energy()
                # print("atoms.info['REF_energy']:", atoms.info["REF_energy"])
            except Exception as e:  # pylint: disable=W0703
                logging.error(f"Failed to extract energy: {e}")
                atoms.info["REF_energy"] = None
    if forces_key == "forces":
        logging.warning(
            "Since ASE version 3.23.0b1, using forces_key 'forces' is no longer safe when communicating between MACE and ASE. We recommend using a different key, rewriting 'forces' to 'REF_forces'. You need to use --forces_key='REF_forces' to specify the chosen key name."
        )
        key_specification.arrays_keys["forces"] = "REF_forces"
        for atoms in atoms_list:
            try:
                atoms.arrays["REF_forces"] = atoms.get_forces()
            except Exception as e:  # pylint: disable=W0703
                logging.error(f"Failed to extract forces: {e}")
                atoms.arrays["REF_forces"] = None
    if stress_key == "stress":
        logging.warning(
            "Since ASE version 3.23.0b1, using stress_key 'stress' is no longer safe when communicating between MACE and ASE. We recommend using a different key, rewriting 'stress' to 'REF_stress'. You need to use --stress_key='REF_stress' to specify the chosen key name."
        )
        key_specification.info_keys["stress"] = "REF_stress"
        for atoms in atoms_list:
            try:
                atoms.info["REF_stress"] = atoms.get_stress()
            except Exception as e:  # pylint: disable=W0703
                atoms.info["REF_stress"] = None

    final_energy_key = key_specification.info_keys["energy"]
    final_forces_key = key_specification.arrays_keys["forces"]
    final_dipole_key = key_specification.info_keys.get("dipole", "REF_dipole")
    has_energy = any(final_energy_key in atoms.info for atoms in atoms_list)
    has_forces = any(final_forces_key in atoms.arrays for atoms in atoms_list)
    has_dipole = any(final_dipole_key in atoms.info for atoms in atoms_list)

    if not has_energy and not has_forces and not has_dipole:
        msg = f"None of '{final_energy_key}', '{final_forces_key}', and '{final_dipole_key}' found in '{file_path}'."
        if no_data_ok:
            logging.warning(msg + " Continuing because no_data_ok=True was passed in.")
        else:
            raise ValueError(
                msg
                + " Please change the key names in the command line arguments or ensure that the file contains the required data."
            )
    if not has_energy:
        logging.warning(
            f"No energies found with key '{final_energy_key}' in '{file_path}'. If this is unexpected, please change the key name in the command line arguments or ensure that the file contains the required data."
        )
    if not has_forces:
        logging.warning(
            f"No forces found with key '{final_forces_key}' in '{file_path}'. If this is unexpected, Please change the key name in the command line arguments or ensure that the file contains the required data."
        )

    if not isinstance(atoms_list, list):
        atoms_list = [atoms_list]

    atomic_energies_dict = {}
    if extract_atomic_energies:
        atoms_without_iso_atoms = []

        for idx, atoms in enumerate(atoms_list):
            atoms.info[head_key] = head_name
            isolated_atom_config = (
                len(atoms) == 1 and atoms.info.get("config_type") == "IsolatedAtom"
            )
            if isolated_atom_config:
                atomic_number = int(atoms.get_atomic_numbers()[0])
                if energy_key in atoms.info.keys():
                    atomic_energies_dict[atomic_number] = float(atoms.info[energy_key])
                else:
                    logging.warning(
                        f"Configuration '{idx}' is marked as 'IsolatedAtom' "
                        "but does not contain an energy. Zero energy will be used."
                    )
                    atomic_energies_dict[atomic_number] = 0.0
            else:
                atoms_without_iso_atoms.append(atoms)

        if len(atomic_energies_dict) > 0:
            logging.info("Using isolated atom energies from training file")
        if not keep_isolated_atoms:
            atoms_list = atoms_without_iso_atoms

    for atoms in atoms_list:
        atoms.info[head_key] = head_name

    configs = config_from_atoms_list(
        atoms_list,
        config_type_weights=config_type_weights,
        key_specification=key_specification,
        head_name=head_name,
    )
    key_specification.info_keys["energy"] = original_energy_key
    key_specification.arrays_keys["forces"] = original_forces_key
    key_specification.info_keys["stress"] = original_stress_key
    return atomic_energies_dict, configs


def compute_average_E0s(
    collections_train: Configurations, z_table: AtomicNumberTable
) -> Dict[int, float]:
    """
    Function to compute the average interaction energy of each chemical element
    returns dictionary of E0s
    """
    len_train = len(collections_train)
    len_zs = len(z_table)
    A = np.zeros((len_train, len_zs))
    B = np.zeros(len_train)
    for i in range(len_train):
        B[i] = collections_train[i].properties["energy"]
        for j, z in enumerate(z_table.zs):
            A[i, j] = np.count_nonzero(collections_train[i].atomic_numbers == z)
    try:
        E0s = np.linalg.lstsq(A, B, rcond=None)[0]
        atomic_energies_dict = {}
        for i, z in enumerate(z_table.zs):
            atomic_energies_dict[z] = E0s[i]
    except np.linalg.LinAlgError:
        logging.error(
            "Failed to compute E0s using least squares regression, using the same for all atoms"
        )
        atomic_energies_dict = {}
        for i, z in enumerate(z_table.zs):
            atomic_energies_dict[z] = 0.0
    return atomic_energies_dict


def save_dataset_as_HDF5(dataset: List, out_name: str) -> None:
    with h5py.File(out_name, "w") as f:
        for i, data in enumerate(dataset):
            save_AtomicData_to_HDF5(data, i, f)


def save_AtomicData_to_HDF5(data, i, h5_file) -> None:
    grp = h5_file.create_group(f"config_{i}")
    grp["num_nodes"] = data.num_nodes
    grp["edge_index"] = data.edge_index
    grp["positions"] = data.positions
    grp["shifts"] = data.shifts
    grp["unit_shifts"] = data.unit_shifts
    grp["cell"] = data.cell
    grp["node_attrs"] = data.node_attrs
    grp["weight"] = data.weight
    grp["energy_weight"] = data.energy_weight
    grp["forces_weight"] = data.forces_weight
    grp["stress_weight"] = data.stress_weight
    grp["virials_weight"] = data.virials_weight
    grp["forces"] = data.forces
    grp["energy"] = data.energy
    grp["stress"] = data.stress
    grp["virials"] = data.virials
    grp["dipole"] = data.dipole
    grp["charges"] = data.charges
    grp["polarizability"] = data.polarizability
    grp["head"] = data.head


def save_configurations_as_HDF5(configurations: Configurations, _, h5_file) -> None:
    grp = h5_file.create_group("config_batch_0")
    for j, config in enumerate(configurations):
        subgroup_name = f"config_{j}"
        subgroup = grp.create_group(subgroup_name)
        subgroup["atomic_numbers"] = write_value(config.atomic_numbers)
        subgroup["positions"] = write_value(config.positions)
        properties_subgrp = subgroup.create_group("properties")
        for key, value in config.properties.items():
            properties_subgrp[key] = write_value(value)
        subgroup["cell"] = write_value(config.cell)
        subgroup["pbc"] = write_value(config.pbc)
        subgroup["weight"] = write_value(config.weight)
        weights_subgrp = subgroup.create_group("property_weights")
        for key, value in config.property_weights.items():
            weights_subgrp[key] = write_value(value)
        subgroup["config_type"] = write_value(config.config_type)


def write_value(value):
    return value if value is not None else "None"



import argparse
import os
import sys
import csv
import numpy as np
import time
import shutil
from pathlib import Path
from tqdm import tqdm

from ase.io import read, write
from ase.io.trajectory import Trajectory
from ase.md.npt import NPT
from ase.md.langevin import Langevin
from ase.build import make_supercell
import ase.units as u

try:
    from ase.md.nose_hoover_chain import NoseHooverChainNVT as NVT_NHC
except ImportError:
    NVT_NHC = None
try:
    from ase.md.nvtberendsen import NVTBerendsen as NVT_Ber
except ImportError:
    NVT_Ber = None
try:
    from ase.md.nptberendsen import NPTBerendsen as NPT_Ber
except ImportError:
    NPT_Ber = None

from ase.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary, ZeroRotation
from ase.geometry import cellpar_to_cell

from macer.calculator.factory import get_calculator, get_available_ffs, ALL_SUPPORTED_FFS
from macer.defaults import DEFAULT_MODELS, DEFAULT_DEVICE, DEFAULT_FF, resolve_model_path, _model_root
from macer.utils.logger import Logger
# from macer import __version__  <-- Moved inside function

# Constants
EV_A3_TO_GPa = 160.21766208
KB_EV_K = 8.617333262145e-5 # Boltzmann constant in eV/K
KJ_MOL_TO_EV = 0.0103642697 # 1 kJ/mol in eV/unit-cell
J_K_MOL_TO_EV_K = 0.0000103642697 # 1 J/mol*K in eV/K/unit-cell

# Global parser storage
_GIBBS_PARSER = None

def get_gibbs_parser(subparsers):
    global _GIBBS_PARSER
    parser = subparsers.add_parser(
        "gibbs",
        help="Calculate Gibbs Free Energy vs Temperature via Temperature Integration (NPT MD)",
        description="Calculate Gibbs Free Energy vs Temperature via Temperature Integration (NPT MD).",
        epilog="""
Examples:
  # 1. Standard Run (Recommended for ScF3 NTE reproduction)
  #    - 2x2x2 Supercell, 100K to 1000K every 50K
  #    - Sufficient sampling (50ps production after 10ps equil)
  macer gibbs -p POSCAR --dim 2 2 2 --temp-start 100 --temp-end 1000 --temp-step 50 --nsteps 50000 --equil-steps 10000

  # 2. Quick Test (Fast check of workflow)
  #    - Coarse grid (200K step), short run (2ps)
  macer gibbs -p POSCAR --dim 2 2 2 --temp-start 100 --temp-end 500 --temp-step 200 --nsteps 2000 --equil-steps 500

  # 3. Absolute Gibbs Energy (Hybrid QHA + MD)
  #    Step A: Generate reference data using Phonopy QHA
  #      macer phonopy qha -p POSCAR --dim 2 2 2 --tmax 300
  #      (This creates 'thermal_properties.yaml' with low-T free energy)
  #
  #    Step B: Run MD integration anchored to the QHA reference
  #      macer gibbs -p POSCAR --dim 2 2 2 --qha-ref thermal_properties.yaml --temp-start 10 --temp-end 1000

  # 4. NVT Mode (Fixed Volume)
  #    - Calculates Helmholtz Free Energy change (F) at fixed volume
  macer gibbs -p POSCAR --ensemble nvt --temp-start 300 --temp-end 1000

  # 5. Gibbs with automatic supercell determination (min length 15A)
  macer gibbs -p POSCAR -l 15.0 --temp-start 100 --temp-end 1000 --ff mattersim
""",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # Input
    input_group = parser.add_argument_group("Input")
    input_group.add_argument("-p", "--poscar", required=False, help="Input structure file (POSCAR or CIF).")
    input_group.add_argument("-f", "--formula", type=str, help="Formula to retrieve from Materials Project (e.g., MgAl2O4)")
    input_group.add_argument("-m", "--mpid", type=str, help="Materials Project ID (e.g., mp-3536)")
    input_group.add_argument("--qha-ref", help="Optional: Path to thermal_properties.yaml (from phonopy) for absolute G reference.")

    # Temperature Scan
    temp_group = parser.add_argument_group("Temperature Scan Settings")
    temp_group.add_argument("--temp-start", type=float, default=100.0, help="Starting temperature (K). Default: 100.0")
    temp_group.add_argument("--temp-end", type=float, default=1000.0, help="Ending temperature (K). Default: 1000.0")
    temp_group.add_argument("--temp-step", type=float, default=50.0, help="Temperature step (K). Default: 50.0")
    temp_group.add_argument("--temps", type=float, nargs='+', help="Specific temperatures to sample. Overrides scan settings.")

    # MD Settings
    md_group = parser.add_argument_group("MD Settings")
    md_group.add_argument("--nsteps", type=int, default=50000, help="Total MD steps per temperature. Default: 50000")
    md_group.add_argument("--equil-steps", type=int, default=10000, help="Equilibration steps to discard per temperature. Default: 10000")
    md_group.add_argument("--tstep", type=float, default=1.0, help="Time step (fs). Default: 1.0")
    md_group.add_argument("--ensemble", choices=["npt", "nvt"], default="npt", help="MD ensemble. Default: npt")
    md_group.add_argument("--press", type=float, default=0.0, help="External pressure (GPa). Default: 0.0")
    md_group.add_argument("--thermostat", choices=["langevin", "nose-hoover", "berendsen"], default="langevin", help="Thermostat algorithm. Default: langevin")
    md_group.add_argument("--ttau", type=float, default=100.0, help="Thermostat time constant (fs). Default: 100.0")
    md_params_group = md_group.add_mutually_exclusive_group() # Placeholder for ptau/friction logic
    md_params_group.add_argument("--ptau", type=float, default=1000.0, help="Barostat time constant (fs). Default: 1000.0")
    md_params_group.add_argument("--friction", type=float, default=1.0, help="Langevin friction (ps^-1). Default: 1.0")
    md_group.add_argument("--dim", type=int, nargs='+', help="Supercell dimension (e.g. '2 2 2').")
    md_group.add_argument("--min-length", "-l", type=float, default=None,
                        help="Minimum supercell lattice vector length in Å to determine DIM automatically if --dim is not set.")

    # MLFF Settings
    mlff_group = parser.add_argument_group("MLFF Model Settings")
    mlff_group.add_argument("--ff", choices=ALL_SUPPORTED_FFS, default=DEFAULT_FF, help="Force field to use.")
    mlff_group.add_argument("--model", help="Path to MLFF model file.")
    mlff_group.add_argument("--device", choices=["cpu", "mps", "cuda"], default=DEFAULT_DEVICE, help=f"Compute device (default: {DEFAULT_DEVICE}).")

    # Output
    out_group = parser.add_argument_group("Output")
    out_group.add_argument("--output-dir", help="Directory for output files.")
    out_group.add_argument("--prefix", default="gibbs", help="Prefix for output files. Default: gibbs")

    parser.set_defaults(func=run_gibbs_workflow)
    _GIBBS_PARSER = parser
    return parser

def run_gibbs_workflow(args):
    from macer.utils.struct_tools import resolve_structure_inputs
    input_list = resolve_structure_inputs(
        poscar_paths=[args.poscar] if args.poscar else None,
        cif_paths=None,
        formula=getattr(args, "formula", None),
        mpid=getattr(args, "mpid", None),
    )
    if not input_list:
        print("Error: Please provide structure input via -p/--poscar, -f/--formula, or -m/--mpid.")
        sys.exit(1)
    args.poscar = input_list[0]

    from macer import __version__
    print(f"--- macer gibbs (v{__version__}): Starting Free Energy Workflow ---")

    # 1. Setup & Directories
    input_path = Path(args.poscar).resolve()
    if args.output_dir:
        output_dir = Path(args.output_dir).resolve()
    else:
        output_dir = input_path.parent / f"gibbs_{input_path.stem}_mlff={args.ff}"
    
    i = 1
    base_output_dir = output_dir
    while output_dir.exists():
        output_dir = Path(f"{base_output_dir}-NEW{i:03d}")
        i += 1
    output_dir.mkdir(parents=True, exist_ok=True)
    
    # Save input structure for record
    shutil.copy(input_path, output_dir / "POSCAR-input")
    
    log_file = output_dir / f"{args.prefix}.log"
    csv_file = output_dir / f"{args.prefix}_results.csv"

    # 2. Temperature List
    if args.temps:
        temp_list = sorted(args.temps)
    else:
        # Use args.temp as start, args.temp_end as end
        start_t = args.temp
        end_t = args.temp_end
        if end_t is None: end_t = 1000.0 # Fallback default
        
        temp_list = np.arange(start_t, end_t + args.temp_step/2, args.temp_step).tolist()
    
    if len(temp_list) < 2:
        print("Error: At least two temperatures are required for integration.")
        sys.exit(1)

    # 3. Reference Energy & Entropy (QHA)
    g_ref_val = None
    s_ref_val = None
    t_ref = temp_list[0]
    
    if args.qha_ref:
        try:
            import yaml
            with open(args.qha_ref, 'r') as f:
                qha_data = yaml.safe_load(f)
            
            # Find closest temperature or interpolate
            qha_temps = np.array([p['temperature'] for p in qha_data['thermal_properties']])
            qha_free_energies = np.array([p['free_energy'] for p in qha_data['thermal_properties']])
            qha_entropies = np.array([p['entropy'] for p in qha_data['thermal_properties']])
            
            # Use linear interpolation for exact matching
            g_ref_raw = float(np.interp(t_ref, qha_temps, qha_free_energies))
            s_ref_raw = float(np.interp(t_ref, qha_temps, qha_entropies))
            
            # Convert units: Phonopy (kJ/mol, J/K/mol) -> Macer (eV, eV/K)
            g_ref_val = g_ref_raw * KJ_MOL_TO_EV
            s_ref_val = s_ref_raw * J_K_MOL_TO_EV_K
            
            print(f"Loaded QHA Reference @ {t_ref} K (Converted from kJ/mol):")
            print(f"  G = {g_ref_val:.6f} eV")
            print(f"  S = {s_ref_val:.6f} eV/K")
            
        except Exception as e:
            print(f"Warning: Failed to load QHA reference from {args.qha_ref}: {e}")

    # 4. Logger Start
    orig_stdout = sys.stdout
    with Logger(str(log_file)) as lg:
        sys.stdout = lg
        print(f"--- Macer Gibbs Free Energy Workflow ---")
        print(f"Command: {' '.join(sys.argv)}")
        print(f"Input POSCAR: {input_path}")
        print(f"Force Field: {args.ff}")
        print(f"Device: {args.device}")
        print(f"Ensemble: {args.ensemble.upper()}")
        print(f"Temperature Range: {temp_list[0]} to {temp_list[-1]} K ({len(temp_list)} steps)")
        print(f"Steps per T: {args.nsteps} (Equil: {args.equil_steps})")

        # 5. Initialize Atoms & Calculator
        atoms = read(str(input_path))
        if args.dim or args.min_length:
            if args.dim:
                if len(args.dim) == 3: sc_matrix = np.diag(args.dim)
                elif len(args.dim) == 9: sc_matrix = np.array(args.dim).reshape(3, 3)
                else: raise ValueError("--dim must be 3 or 9 integers.")
                print(f"Using user-provided supercell DIM = {args.dim}")
            else:
                # Auto-determine from min_length
                initial_cell = atoms.get_cell()
                scaling_factors_list = [max(1, int(np.ceil(args.min_length / np.linalg.norm(v)))) for v in initial_cell]
                sc_matrix = np.diag(scaling_factors_list)
                args.dim = scaling_factors_list # Update args.dim for scaling logic later
                print(f"Auto-determining supercell from --min-length. DIM = {scaling_factors_list}")
            
            atoms = make_supercell(atoms, sc_matrix)
            print(f"Applied Supercell (Atoms: {len(atoms)})")
            # Save SPOSCAR
            write(str(output_dir / "SPOSCAR"), atoms, format='vasp')
        else:
            # If no dim or min_length, input is SPOSCAR
            write(str(output_dir / "SPOSCAR"), atoms, format='vasp')
        
        # Upper triangular cell for NPT
        tri_cell = cellpar_to_cell(atoms.cell.cellpar())
        atoms.set_cell(tri_cell, scale_atoms=True)
        atoms.pbc = True

        # Species for XDATCAR
        symbols = atoms.get_chemical_symbols()
        species = list(dict.fromkeys(symbols))
        counts = [symbols.count(s) for s in species]

        # Calculator
        model_path = args.model
        if model_path is None:
            default_model_name = DEFAULT_MODELS.get(args.ff)
            if default_model_name:
                model_path = resolve_model_path(default_model_name)
        else:
            model_path = resolve_model_path(model_path)
            
        calc_kwargs = {"device": args.device}
        if args.ff == "mace": calc_kwargs["model_paths"] = [model_path]
        else: calc_kwargs["model_path"] = model_path
        
        calculator = get_calculator(ff_name=args.ff, **calc_kwargs)
        atoms.calc = calculator

        # 6. Data Loop
        results = []
        n_atoms = len(atoms)
        
        for idx, target_temp in enumerate(temp_list):
            print(f"\n>>> Temperature Step: {target_temp} K ({idx+1}/{len(temp_list)})")
            
            # Setup specific output files for this temperature
            xdatcar_path = output_dir / f"XDATCAR-{target_temp:.0f}K"
            md_log_path = output_dir / f"md-{target_temp:.0f}K.log"
            
            xdat_f = open(xdatcar_path, 'w')
            md_log_f = open(md_log_path, 'w')
            
            # MD Log Header
            md_log_f.write("Step      Time[ps]      Etot[eV]     Epot[eV]     Ekin[eV]    T[K]      a[A]       b[A]       c[A]      Vol[A^3]\n")

            # Initial velocities
            if idx == 0:
                MaxwellBoltzmannDistribution(atoms, temperature_K=target_temp, force_temp=True)
                Stationary(atoms); ZeroRotation(atoms)
            
            # Setup Integrator
            timestep = args.tstep * u.fs
            
            # Handle defaults (cli.py sets 0 for auto, gibbs defaults were 100/1000)
            ttau_val = args.ttau if args.ttau > 0 else 100.0
            ptau_val = args.ptau if args.ptau > 0 else 1000.0
            
            ttime = ttau_val * u.fs
            
            if args.ensemble == "npt":
                extstress = args.press * u.GPa
                # Standard ASE NPT pfactor calculation:
                # pfactor should be in units of (time)^2 * (energy)/(volume)
                # We use a robust default bulk modulus (~75 GPa) for the estimate
                bulk_modulus = 75.0 * u.GPa
                pfact = (ptau_val * u.fs)**2 * bulk_modulus
                
                # Explicitly set mask to avoid potential TypeError with some ASE versions (bool * NoneType)
                mask = np.ones((3, 3))

                if args.thermostat == "langevin":
                    effective_tau = 1000.0 / args.friction if args.friction > 0 else ttau_val
                    ttime = effective_tau * u.fs
                    # Note: ASE NPT uses 'ttime' for thermostat damping
                    dyn = NPT(atoms, timestep=timestep, temperature_K=target_temp,
                              externalstress=extstress, ttime=ttime, pfactor=pfact, mask=mask)
                else:
                    dyn = NPT(atoms, timestep=timestep, temperature_K=target_temp,
                              externalstress=extstress, ttime=ttime, pfactor=pfact, mask=mask)
            else: # NVT
                if args.thermostat == "langevin":
                    fric_val = (args.friction * 1e-3) if args.friction else (1.0 / args.ttau)
                    dyn = Langevin(atoms, timestep=timestep, temperature_K=target_temp, friction=fric_val)
                elif args.thermostat == "berendsen" and NVT_Ber:
                    dyn = NVT_Ber(atoms, timestep=timestep, temperature_K=target_temp, taut=ttime)
                else:
                    if NVT_NHC:
                        dyn = NVT_NHC(atoms, timestep=timestep, temperature_K=target_temp, tdamp=ttime)
                    else:
                        dyn = NVT_Ber(atoms, timestep=timestep, temperature_K=target_temp, taut=ttime)

            # Sampling holders
            epot_vals = []
            vol_vals = []
            a_vals, b_vals, c_vals = [], [], []
            step_counter = 0

            def sample():
                nonlocal step_counter
                step_counter += 1
                
                epot = atoms.get_potential_energy()
                ekin = atoms.get_kinetic_energy()
                etot = epot + ekin
                temp = atoms.get_temperature()
                vol = atoms.get_volume()
                abc = atoms.cell.lengths()
                
                epot_vals.append(epot)
                vol_vals.append(vol)
                a_vals.append(abc[0])
                b_vals.append(abc[1])
                c_vals.append(abc[2])
                
                # Write to MD Log
                time_ps = step_counter * args.tstep / 1000.0
                md_log_f.write(f"{step_counter:<10d} {time_ps:<12.4f} {etot:<12.4f} {epot:<12.4f} {ekin:<11.4f} {temp:<9.2f} {abc[0]:<10.4f} {abc[1]:<10.4f} {abc[2]:<10.4f} {vol:<10.4f}\n")
                
                # Write to XDATCAR
                if step_counter == 1:
                    xdat_f.write(" ".join(species) + "\n")
                    xdat_f.write("    1.000000\n")
                    for vec in atoms.cell: xdat_f.write(f"     {vec[0]:11.6f} {vec[1]:11.6f} {vec[2]:11.6f}\n")
                    xdat_f.write(" " + " ".join(f"{s:>2}" for s in species) + "\n")
                    xdat_f.write(" " + " ".join(f"{c:16d}" for c in counts) + "\n")
                
                if step_counter > 1 and args.ensemble == "npt":
                    xdat_f.write(" ".join(species) + "\n")
                    xdat_f.write("    1.000000\n")
                    for vec in atoms.cell: xdat_f.write(f"     {vec[0]:11.6f} {vec[1]:11.6f} {vec[2]:11.6f}\n")
                    xdat_f.write(" " + " ".join(f"{s:>2}" for s in species) + "\n")
                    xdat_f.write(" " + " ".join(f"{c:16d}" for c in counts) + "\n")
                
                xdat_f.write(f"Direct configuration= {step_counter:5d}\n")
                for s in atoms.get_scaled_positions(wrap=True):
                    xdat_f.write(f"   {s[0]:.8f}   {s[1]:.8f}   {s[2]:.8f}\n")

            # Helper for progress bar
            def run_with_progress(dynamics, steps, desc):
                pbar = tqdm(total=steps, desc=desc, ncols=80)
                def update_pbar():
                    pbar.update(1)
                dynamics.attach(update_pbar, interval=1)
                dynamics.run(steps)
                pbar.close()

            # 1. Equilibration
            if args.equil_steps > 0:
                run_with_progress(dyn, args.equil_steps, f"  Equilibration ({args.equil_steps})")
            
            # 2. Production
            dyn.attach(sample, interval=1)
            run_with_progress(dyn, args.nsteps, f"  Production    ({args.nsteps})")

            # Close temp files
            xdat_f.close()
            md_log_f.close()

            # Statistics
            avg_epot = np.mean(epot_vals)
            avg_vol = np.mean(vol_vals)
            avg_a, avg_b, avg_c = np.mean(a_vals), np.mean(b_vals), np.mean(c_vals)
            
            avg_ekin = 1.5 * n_atoms * KB_EV_K * target_temp
            pv_term = (args.press * avg_vol) / EV_A3_TO_GPa
            avg_h = avg_epot + avg_ekin + pv_term
            
            std_h = np.std(epot_vals) / np.sqrt(len(epot_vals)) 
            
            results.append({
                'T_K': target_temp,
                'H_avg_eV': avg_h,
                'H_err_eV': std_h,
                'E_pot_avg_eV': avg_epot,
                'V_avg_A3': avg_vol,
                'a_avg_A': avg_a,
                'b_avg_A': avg_b,
                'c_avg_A': avg_c
            })
            
            print(f"  Results: H={avg_h:.4f} eV, V={avg_vol:.2f} A^3, a={avg_a:.4f} A")

        # 7. Integration
        print("\n--- Performing Integration ---")
        temps = np.array([r['T_K'] for r in results])
        enthalpies = np.array([r['H_avg_eV'] for r in results])
        integrands = -enthalpies / (temps**2)
        
        # NumPy 2.0 compatibility for trapezoidal integration
        try:
            trapz = np.trapezoid
        except AttributeError:
            trapz = np.trapz

        integrals = []
        for i in range(len(temps)):
            if i == 0: integrals.append(0.0)
            else:
                val = trapz(integrands[:i+1], temps[:i+1])
                integrals.append(val)
        
        integrals = np.array(integrals)
        delta_g_rel = temps * integrals
        
        g_abs = [None] * len(temps)
        s_abs = [None] * len(temps)
        
        if g_ref_val is not None:
             g_t0_ref = g_ref_val
             if args.dim:
                 scale = np.prod(args.dim)
                 g_t0_ref *= scale
                 # If s_ref_val exists, scale it too
                 if s_ref_val: s_ref_val *= scale
                 print(f"  Scaling Reference by {scale} (Supercell count)")
             
             # Calculate G_abs
             for i in range(len(temps)):
                 g_abs[i] = temps[i] * (g_t0_ref / temps[0] + integrals[i])
             
             # Calculate S_abs = -dG/dT
             # Use gradient for interior, but handle boundaries carefully
             # If we have enough points, np.gradient is good.
             if len(temps) > 1:
                 # G_abs is list, convert to np array
                 g_arr = np.array(g_abs)
                 # Calculate S = -dG/dT
                 s_grad = -np.gradient(g_arr, temps)
                 s_abs = s_grad.tolist()
                 
                 # Fix the first point to match S_ref if available (more accurate than 1-sided diff)
                 if s_ref_val is not None:
                     # Blend or replace? Let's replace for consistency at T0
                     s_abs[0] = s_ref_val
        
        # 8. Save CSV
        print(f"Saving results to {csv_file}")
        EV_K_TO_J_MOL_K = 96485.33212
        with open(csv_file, 'w', newline='') as f:
            writer = csv.writer(f)
            header = ["T_K", "H_avg_eV", "H_err_eV", "V_avg_A3", "a_avg_A", "b_avg_A", "c_avg_A", 
                      "E_pot_avg_eV", "Integrand", "Integral_val", "Delta_G_rel_eV", "G_abs_eV", 
                      "S_grad_eV/K", "S_thermo_eV/K",
                      "Cp_JmolK", "alpha_MK"]
            writer.writerow(header)
            
            for i, res in enumerate(results):
                t = res['T_K']
                h = res['H_avg_eV']
                g_val = g_abs[i] if g_abs[i] is not None else 0.0
                
                # S from (H-G)/T
                s_thermo = (h - g_val) / t if g_abs[i] is not None and t > 0 else 0.0
                
                # S from gradient
                s_grad_val = s_abs[i] if s_abs[i] is not None else 0.0
                
                # Calculate derived properties for CSV (Backward difference)
                cp_val = ""
                al_val = ""
                if i > 0:
                    t_prev = results[i-1]['T_K']
                    h_prev = results[i-1]['H_avg_eV']
                    a_prev = results[i-1]['a_avg_A']
                    dt = t - t_prev
                    
                    if dt > 0:
                        cp_sys = (h - h_prev) / dt
                        cp_val = (cp_sys / n_atoms) * EV_K_TO_J_MOL_K
                        
                        a_mean = (res['a_avg_A'] + a_prev) / 2.0
                        al_val = (1.0 / a_mean) * ((res['a_avg_A'] - a_prev) / dt) * 1e6

                row = [
                    t, h, res['H_err_eV'], res['V_avg_A3'], res['a_avg_A'], res['b_avg_A'], res['c_avg_A'],
                    res['E_pot_avg_eV'], integrands[i], integrals[i], delta_g_rel[i], 
                    g_abs[i] if g_abs[i] is not None else "",
                    s_grad_val if s_abs[i] is not None else "",
                    s_thermo if g_abs[i] is not None else "",
                    cp_val, al_val
                ]
                writer.writerow(row)

        print("\n--- Workflow Completed Successfully ---")
        print(f"Log: {log_file}")
        print(f"CSV: {csv_file}")

        # 9. Summary Table Output
        print("\n" + "="*115)
        print("SUMMARY of GIBBS FREE ENERGY INTEGRATION")
        print("="*115)
        header = f"{'T (K)':>8} | {'H (eV)':>10} | {'Vol (A3)':>10} | {'G_abs (eV)':>12} | {'S (eV/K)':>12} | {'TS (eV)':>12}"
        print(header)
        print("-" * 115)
        
        for i, res in enumerate(results):
            t = res['T_K']
            h = res['H_avg_eV']
            v = res['V_avg_A3']
            
            ga = g_abs[i] if g_abs[i] is not None else float('nan')
            sa = s_abs[i] if s_abs[i] is not None else float('nan')
            
            ga_str = f"{ga:12.4f}" if not np.isnan(ga) else f"{'N/A':>12}"
            sa_str = f"{sa:12.6f}" if not np.isnan(sa) else f"{'N/A':>12}"
            
            ts_val = t * sa if not np.isnan(sa) else float('nan')
            ts_str = f"{ts_val:12.4f}" if not np.isnan(ts_val) else f"{'N/A':>12}"
            
            print(f"{t:8.1f} | {h:10.4f} | {v:10.4f} | {ga_str} | {sa_str} | {ts_str}")
            
        print("="*115)
        print(" * S is calculated via numerical differentiation (-dG/dT).")
        print(" * Check CSV for more details (including (H-G)/T based Entropy).")

        # 10. Derived Properties (Cp, alpha)
        if len(results) >= 2:
            print("\n" + "-"*65)
            print("DERIVED PROPERTIES (Approximate from finite differences)")
            print("-"*65)
            print(f"{'T_mid (K)':>10} | {'Cp (J/molK)':>15} | {'alpha (10^-6/K)':>18}")
            print("-"*65)
            
            # Constants for conversion
            # 1 eV/K = 96485.332 J/mol*K (Faraday constant approx)
            EV_K_TO_J_MOL_K = 96485.33212
            
            for i in range(len(results) - 1):
                t1, t2 = results[i]['T_K'], results[i+1]['T_K']
                h1, h2 = results[i]['H_avg_eV'], results[i+1]['H_avg_eV']
                a1, a2 = results[i]['a_avg_A'], results[i+1]['a_avg_A']
                
                dt = t2 - t1
                t_mid = (t1 + t2) / 2.0
                
                # Heat Capacity: Cp = dH/dT (normalized per mole-atom)
                
                cp_system_ev_k = (h2 - h1) / dt
                cp_per_atom_ev_k = cp_system_ev_k / n_atoms
                cp_molar_j_mol_k = cp_per_atom_ev_k * EV_K_TO_J_MOL_K
                
                # Thermal Expansion: alpha = (1/a) * da/dT
                a_mean = (a1 + a2) / 2.0
                alpha_val = (1.0 / a_mean) * ((a2 - a1) / dt)
                alpha_mk = alpha_val * 1e6 # in 10^-6 K^-1
                
                print(f"{t_mid:10.1f} | {cp_molar_j_mol_k:15.3f} | {alpha_mk:18.3f}")
            print("-"*65)
            print(" * Cp is normalized per mole-atom (J/mol-atom*K).")
            print(" * Alpha is linear expansion coefficient.")
            print("="*65 + "\n")

    sys.stdout = orig_stdout



import macer.externals.dynaphopy_bundled.dynamics as dyn
import numpy as np
from macer.externals.dynaphopy_bundled.power_spectrum import _progress_bar
from lammps import lammps
from macer.externals.dynaphopy_bundled.interface.iofile import get_correct_arrangement


def generate_lammps_trajectory(structure,
                               input_file,
                               total_time=0.1,  # picoseconds
                               time_step=0.002,  # picoseconds
                               relaxation_time=0,
                               silent=False,
                               supercell=(1, 1, 1),
                               memmap=False,  # not fully implemented yet!
                               velocity_only=False,
                               lammps_log=True,
                               temperature=None,
                               thermostat_mass=0.5,
                               sampling_interval=1):  # in timesteps

    cmdargs_lammps = ['-echo','none', '-screen', 'none']
    if not lammps_log:
        cmdargs_lammps += ['-log', 'none']

    lmp = lammps(cmdargs=cmdargs_lammps)

    lmp.file(input_file)
    lmp.command('timestep {}'.format(time_step))

    lmp.command('replicate {} {} {}'.format(*supercell))

    # Force reset temperature (overwrites lammps script)
    # This forces NVT simulation
    if temperature is not None:
        print ('Temperature reset to {} (NVT)'.format(temperature))
        lmp.command('fix             int all nvt temp {0} {0} {1}'.format(temperature,
                                                                          thermostat_mass))
    # Check if correct number of atoms
    if lmp.extract_global("natoms",0) < 2:
        print ('Number of atoms in MD should be higher than 1!')
        exit()

    # Check if initial velocities all zero
    if not np.array(lmp.gather_atoms("v", 1, 3)).any():
        print('Set lammps initial velocities')
        t = temperature if temperature is not None else 100
        lmp.command('velocity        all create {} 3627941 dist gaussian mom yes'.format(t))
        lmp.command('velocity        all scale {}'.format(t))

    lmp.command('run 0')

    # natoms = lmp.extract_global("natoms",0)
    # mass = lmp.extract_atom("mass",2)
    # temp = lmp.extract_compute("thermo_temp",0,0)
    # print("Temperature from compute =",temp)
    # print("Natoms, mass, x[0][0] coord =", natoms, mass[1], x[0][0])
    # print ('thermo', lmp.get_thermo('1'))

    try:
        xlo =lmp.extract_global("boxxlo")
        xhi =lmp.extract_global("boxxhi")
        ylo =lmp.extract_global("boxylo")
        yhi =lmp.extract_global("boxyhi")
        zlo =lmp.extract_global("boxzlo")
        zhi =lmp.extract_global("boxzhi")
        xy =lmp.extract_global("xy")
        yz =lmp.extract_global("yz")
        xz =lmp.extract_global("xz")

    except TypeError:
        xlo =lmp.extract_global("boxxlo", 1)
        xhi =lmp.extract_global("boxxhi", 1)
        ylo =lmp.extract_global("boxylo", 1)
        yhi =lmp.extract_global("boxyhi", 1)
        zlo =lmp.extract_global("boxzlo", 1)
        zhi =lmp.extract_global("boxzhi", 1)
        xy =lmp.extract_global("xy", 1)
        yz =lmp.extract_global("yz", 1)
        xz =lmp.extract_global("xz", 1)

    except UnboundLocalError:
        boxlo, boxhi, xy, yz, xz, periodicity, box_change = lmp.extract_box()
        xlo, ylo, zlo = boxlo
        xhi, yhi, zhi = boxhi

    simulation_cell = np.array([[xhi-xlo, xy,  xz],
                           [0,  yhi-ylo,  yz],
                           [0,   0,  zhi-zlo]]).T

    positions = []
    velocity = []
    energy = []

    na = lmp.get_natoms()
    id = lmp.extract_atom("id", 0)
    id = np.array([id[i] - 1 for i in range(na)], dtype=int)

    xp = lmp.extract_atom("x", 3)
    reference = np.array([[xp[i][0], xp[i][1], xp[i][2]] for i in range(na)], dtype=float)[id, :]

    # xc = lmp.gather_atoms("x", 1, 3)
    # reference = np.array([xc[i] for i in range(na*3)]).reshape((na,3))

    template = get_correct_arrangement(reference, structure)
    indexing = np.argsort(template)

    lmp.command('variable energy equal pe'.format(int(relaxation_time/time_step)))
    lmp.command('run {}'.format(int(relaxation_time/time_step)))

    if not silent:
        _progress_bar(0, 'lammps')

    n_loops = int(total_time / time_step / sampling_interval)
    for i in range(n_loops):
        if not silent:
            _progress_bar(float((i+1) * time_step * sampling_interval) / total_time, 'lammps', )

        lmp.command('run {}'.format(sampling_interval))

        # xc = lmp.gather_atoms("x", 1, 3)
        # vc = lmp.gather_atoms("v", 1, 3)
        energy.append(lmp.extract_variable('energy', 'all', 0))

        id = lmp.extract_atom("id", 0)
        id = np.array([id[i] - 1 for i in range(na)], dtype=int)

        vc = lmp.extract_atom("v", 3)
        velocity.append(np.array([[vc[i][0], vc[i][1], vc[i][2]] for i in range(na)], dtype=float)[id, :][indexing, :])
        # velocity.append(np.array([vc[i] for i in range(na * 3)]).reshape((na, 3))[indexing, :])

        if not velocity_only:
            xc = lmp.extract_atom("x", 3)
            positions.append(np.array([[xc[i][0], xc[i][1], xc[i][2]] for i in range(na)], dtype=float)[id, :][indexing, :])
            # positions.append(np.array([xc[i] for i in range(na * 3)]).reshape((na, 3))[indexing, :])

    positions = np.array(positions, dtype=complex)
    velocity = np.array(velocity, dtype=complex)
    energy = np.array(energy)

    if velocity_only:
        positions = None

    lmp.close()

    time = np.array([i * time_step * sampling_interval for i in range(n_loops)], dtype=float)


    return dyn.Dynamics(structure=structure,
                        trajectory=positions,
                        velocity=velocity,
                        time=time,
                        energy=energy,
                        supercell=simulation_cell,
                        memmap=memmap)


if __name__ == '__main__':

    structure = None
    print (generate_lammps_trajectory(structure, 'in.demo'))


"""
Wrapper class for o3.Linear that optionally uses cuet.Linear
"""

import dataclasses
import types
from typing import List, Optional

import torch
from macer.externals.e3nn_legacy import o3

from macer.externals.mace_bundled.modules.symmetric_contraction import SymmetricContraction
from macer.externals.mace_bundled.tools.cg import O3_e3nn
from macer.externals.mace_bundled.tools.scatter import scatter_sum

try:
    import cuequivariance as cue
    import cuequivariance_torch as cuet

    CUET_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    CUET_AVAILABLE = False

try:
    import openequivariance as oeq

    OEQ_AVAILABLE = True
except ImportError:
    OEQ_AVAILABLE = False


@dataclasses.dataclass
class CuEquivarianceConfig:
    """Configuration for cuequivariance acceleration"""

    enabled: bool = False
    layout: str = "mul_ir"  # One of: mul_ir, ir_mul
    layout_str: str = "mul_ir"
    group: str = "O3"
    optimize_all: bool = False  # Set to True to enable all optimizations
    optimize_linear: bool = False
    optimize_channelwise: bool = False
    optimize_symmetric: bool = False
    optimize_fctp: bool = False
    conv_fusion: bool = False  # Set to True to enable conv fusion

    def __post_init__(self):
        if self.enabled and CUET_AVAILABLE:
            self.layout_str = self.layout
            self.layout = getattr(cue, self.layout)
            self.group = (
                O3_e3nn if self.group == "O3_e3nn" else getattr(cue, self.group)
            )
        if not CUET_AVAILABLE:
            self.enabled = False


@dataclasses.dataclass
class OEQConfig:
    """Configuration for cuequivariance acceleration"""

    enabled: bool = False
    optimize_all: bool = False
    optimize_channelwise: bool = False
    conv_fusion: Optional[str] = "atomic"

    def __post_init__(self):
        if not OEQ_AVAILABLE:
            self.enabled = False


class Linear:
    """Returns either a cuet.Linear or o3.Linear based on config"""

    def __new__(
        cls,
        irreps_in: o3.Irreps,
        irreps_out: o3.Irreps,
        shared_weights: bool = True,
        internal_weights: bool = True,
        cueq_config: Optional[CuEquivarianceConfig] = None,
    ):
        if (
            CUET_AVAILABLE
            and cueq_config is not None
            and cueq_config.enabled
            and (cueq_config.optimize_all or cueq_config.optimize_linear)
        ):
            return cuet.Linear(
                cue.Irreps(cueq_config.group, irreps_in),
                cue.Irreps(cueq_config.group, irreps_out),
                layout=cueq_config.layout,
                shared_weights=shared_weights,
                use_fallback=True,
            )

        return o3.Linear(
            irreps_in,
            irreps_out,
            shared_weights=shared_weights,
            internal_weights=internal_weights,
        )


def with_scatter_sum(conv_tp: torch.nn.Module) -> torch.nn.Module:
    conv_tp.original_forward = conv_tp.forward

    def forward(
        self,
        node_feats: torch.Tensor,
        edge_attrs: torch.Tensor,
        tp_weights: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> torch.Tensor:
        sender = edge_index[0]
        receiver = edge_index[1]
        num_nodes = node_feats.shape[0]

        mji = self.original_forward(node_feats[sender], edge_attrs, tp_weights)
        message = scatter_sum(src=mji, index=receiver, dim=0, dim_size=num_nodes)
        return message

    conv_tp.forward = types.MethodType(forward, conv_tp)
    return conv_tp


def with_cueq_conv_fusion(conv_tp: torch.nn.Module) -> torch.nn.Module:
    """Wraps a cuet.ConvTensorProduct to use conv fusion"""
    conv_tp.original_forward = conv_tp.forward
    num_segment = conv_tp.m.buffer_num_segments[0]
    num_operands = conv_tp.m.operand_extent
    conv_tp.weight_numel = num_segment * num_operands

    def forward(
        self,
        node_feats: torch.Tensor,
        edge_attrs: torch.Tensor,
        tp_weights: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> torch.Tensor:
        sender = edge_index[0]
        receiver = edge_index[1]
        return self.original_forward(
            [tp_weights, node_feats, edge_attrs],
            {1: sender},
            {0: node_feats},
            {0: receiver},
        )[0]

    conv_tp.forward = types.MethodType(forward, conv_tp)
    return conv_tp


class TensorProduct:
    """Wrapper around o3.TensorProduct/cuet.ChannelwiseTensorProduct/oeq.TensorProduct followed by a scatter sum"""

    def __new__(
        cls,
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irreps,
        irreps_out: o3.Irreps,
        instructions: Optional[List] = None,
        shared_weights: bool = False,
        internal_weights: bool = False,
        cueq_config: Optional[CuEquivarianceConfig] = None,
        oeq_config: Optional[OEQConfig] = None,
    ):
        if (
            CUET_AVAILABLE
            and cueq_config is not None
            and cueq_config.enabled
            and (cueq_config.optimize_all or cueq_config.optimize_channelwise)
        ):
            if cueq_config.conv_fusion:
                return with_cueq_conv_fusion(
                    cuet.SegmentedPolynomial(
                        cue.descriptors.channelwise_tensor_product(
                            cue.Irreps(cueq_config.group, irreps_in1),
                            cue.Irreps(cueq_config.group, irreps_in2),
                            cue.Irreps(cueq_config.group, irreps_out),
                        )
                        .flatten_coefficient_modes()
                        .squeeze_modes()
                        .polynomial,
                        math_dtype=torch.get_default_dtype(),
                    )
                )
            return cuet.ChannelWiseTensorProduct(
                cue.Irreps(cueq_config.group, irreps_in1),
                cue.Irreps(cueq_config.group, irreps_in2),
                cue.Irreps(cueq_config.group, irreps_out),
                layout=cueq_config.layout,
                shared_weights=shared_weights,
                internal_weights=internal_weights,
                dtype=torch.get_default_dtype(),
                math_dtype=torch.get_default_dtype(),
            )
        if (
            OEQ_AVAILABLE
            and oeq_config is not None
            and oeq_config.enabled
            and (oeq_config.optimize_all or oeq_config.optimize_channelwise)
        ):
            dtype = oeq.torch_to_oeq_dtype(torch.get_default_dtype())
            tpp = oeq.TPProblem(
                irreps_in1,
                irreps_in2,
                irreps_out,
                instructions,
                shared_weights=shared_weights,
                internal_weights=internal_weights,
                irrep_dtype=dtype,
                weight_dtype=dtype,
            )

            if oeq_config.conv_fusion is None:
                return oeq.TensorProduct(tpp)
            if oeq_config.conv_fusion == "atomic":
                return oeq.TensorProductConv(tpp, deterministic=False)

            raise ValueError(f"Unknown conv_fusion option: {oeq_config.conv_fusion}")

        return o3.TensorProduct(
            irreps_in1,
            irreps_in2,
            irreps_out,
            instructions=instructions,
            shared_weights=shared_weights,
            internal_weights=internal_weights,
        )


class FullyConnectedTensorProduct:
    """Wrapper around o3.FullyConnectedTensorProduct/cuet.FullyConnectedTensorProduct"""

    def __new__(
        cls,
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irreps,
        irreps_out: o3.Irreps,
        shared_weights: bool = True,
        internal_weights: bool = True,
        cueq_config: Optional[CuEquivarianceConfig] = None,
    ):
        if (
            CUET_AVAILABLE
            and cueq_config is not None
            and cueq_config.enabled
            and (cueq_config.optimize_all or cueq_config.optimize_fctp)
        ):
            return cuet.FullyConnectedTensorProduct(
                cue.Irreps(cueq_config.group, irreps_in1),
                cue.Irreps(cueq_config.group, irreps_in2),
                cue.Irreps(cueq_config.group, irreps_out),
                layout=cueq_config.layout,
                shared_weights=shared_weights,
                internal_weights=internal_weights,
                use_fallback=True,
            )

        return o3.FullyConnectedTensorProduct(
            irreps_in1,
            irreps_in2,
            irreps_out,
            shared_weights=shared_weights,
            internal_weights=internal_weights,
        )


class SymmetricContractionWrapper:
    """Wrapper around SymmetricContraction/cuet.SymmetricContraction"""

    def __new__(
        cls,
        irreps_in: o3.Irreps,
        irreps_out: o3.Irreps,
        correlation: int,
        num_elements: Optional[int] = None,
        cueq_config: Optional[CuEquivarianceConfig] = None,
        oeq_config: Optional[OEQConfig] = None,  # pylint: disable=unused-argument
        use_reduced_cg: bool = True,
    ):
        use_reduced_cg = use_reduced_cg and CUET_AVAILABLE
        if (
            CUET_AVAILABLE
            and cueq_config is not None
            and cueq_config.enabled
            and (cueq_config.optimize_all or cueq_config.optimize_symmetric)
        ):
            return cuet.SymmetricContraction(
                cue.Irreps(cueq_config.group, irreps_in),
                cue.Irreps(cueq_config.group, irreps_out),
                layout_in=cue.ir_mul,
                layout_out=cueq_config.layout,
                contraction_degree=correlation,
                num_elements=num_elements,
                original_mace=(not use_reduced_cg),
                dtype=torch.get_default_dtype(),
                math_dtype=torch.get_default_dtype(),
            )

        return SymmetricContraction(
            irreps_in=irreps_in,
            irreps_out=irreps_out,
            correlation=correlation,
            num_elements=num_elements,
            use_reduced_cg=use_reduced_cg,
        )


class TransposeIrrepsLayoutWrapper:
    """Wrapper around cuet.TransposeIrrepsLayout"""

    def __new__(
        cls,
        irreps: o3.Irreps,
        source: str,
        target: str,
        cueq_config: Optional[CuEquivarianceConfig] = None,
    ):
        if CUET_AVAILABLE and cueq_config is not None and cueq_config.enabled:
            return cuet.TransposeIrrepsLayout(
                cue.Irreps(cueq_config.group, irreps),
                source=getattr(cue, source),
                target=getattr(cue, target),
                use_fallback=True,
            )

        return None



import numpy as np
import sys
from macer.externals.dynaphopy_bundled.c.displacements import atomic_displacements


def progress_bar(progress):
    bar_length = 30
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "Progress error\r\n"
    if progress < 0:
        progress = 0
        status = "Halt ...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(bar_length*progress))
    text = "\rTrajectory: [{0}] {1:.2f}% {2}".format("#"*block + "-"*(bar_length-block),
                                                      progress*100, status)
    sys.stdout.write(text)
    sys.stdout.flush()


# Not used (only for test)
def relativize_trajectory(dynamic, memmap=False):

    cell = dynamic.get_supercell()
    number_of_atoms = dynamic.trajectory.shape[1]
    supercell = dynamic.get_supercell_matrix()
    position = dynamic.structure.get_positions(supercell=supercell)
    # normalized_trajectory = np.zeros_like(dynamic.trajectory.real)
    normalized_trajectory = dynamic.trajectory.copy()

    trajectory = dynamic.trajectory

    if memmap:
        normalized_trajectory = np.memmap('r_trajectory.map', dtype='complex', mode='w+', shape=trajectory.shape)
    else:
        normalized_trajectory = dynamic.trajectory.copy()

    # progress_bar(0)

    for i in range(number_of_atoms):
        normalized_trajectory[:, i, :] = atomic_displacements(trajectory[:, i, :], position[i], cell)

   # progress_bar(float(i+1)/number_of_atoms)
    return normalized_trajectory

# Not used (only for test)
def relativize_trajectory_py(dynamic):

    print('Using python rutine for calculating atomic displacements')
    cell = dynamic.get_supercell()
    number_of_atoms = dynamic.trajectory.shape[1]
    supercell = dynamic.get_supercell_matrix()
    position = dynamic.structure.get_positions(supercell=supercell)
    normalized_trajectory = dynamic.trajectory.real.copy()

    progress_bar(0)

    for j in range(number_of_atoms):
        for i in range(0, normalized_trajectory.shape[0]):

            difference = normalized_trajectory[i, j, :] - position[j]

            # difference_matrix = np.array(np.dot(np.linalg.inv(cell),(IniSep)),dtype=int)
            difference_matrix = np.around(np.dot(np.linalg.inv(cell), difference), decimals=0)
            normalized_trajectory[i, j, :] -= np.dot(difference_matrix, cell) + position[j]

        progress_bar(float(j+1)/number_of_atoms)

    return normalized_trajectory


def trajectory_projection(dynamic, direction):

    direction = np.array(direction)/np.linalg.norm(direction)

    supercell = dynamic.get_supercell_matrix()
    trajectory = dynamic.get_relative_trajectory()
    atom_type_index = dynamic.structure.get_atom_type_index(supercell=supercell)
    number_of_atom_types = dynamic.structure.get_number_of_atom_types()

    projections = []

    for j in range(number_of_atom_types):
        projection = np.array([])
        for i in range(0, trajectory.shape[1]):
            if atom_type_index[i] == j:
                # print('atom:', i, 'type:', atom_type_index[i])
                projection = np.append(projection, np.dot(trajectory[:, i, :].real,
                                                          direction/np.linalg.norm(direction)))
        projections.append(projection)

    return np.array(projections)



import os
import sys
import types
import importlib.metadata as _importlib_metadata
from macer.defaults import DEFAULT_MODELS, _macer_root, resolve_model_path
from macer.utils.model_manager import ensure_model, print_model_help

def _ensure_pkg_resources():
    try:
        import pkg_resources  # noqa: F401
        return
    except ModuleNotFoundError:
        pass

    class DistributionNotFound(Exception):
        pass

    def get_distribution(dist_name):
        try:
            version = _importlib_metadata.version(dist_name)
        except _importlib_metadata.PackageNotFoundError as e:
            raise DistributionNotFound(str(e))
        return types.SimpleNamespace(version=version)

    shim = types.ModuleType("pkg_resources")
    shim.get_distribution = get_distribution
    shim.DistributionNotFound = DistributionNotFound
    sys.modules["pkg_resources"] = shim

_ensure_pkg_resources()

try:
    from mattersim.forcefield import MatterSimCalculator
    _MATTERSIM_AVAILABLE = True
except Exception as e:
    import traceback
    # print(f"DEBUG: MatterSim import failed: {e}")
    # traceback.print_exc()
    _MATTERSIM_ERR = f"{e}\n{traceback.format_exc()}"
    _MATTERSIM_AVAILABLE = False
    MatterSimCalculator = None  # Define as None if import fails

def get_mattersim_calculator(model_path, device="cpu", **kwargs):
    if not _MATTERSIM_AVAILABLE:
        print(f"ERROR DETAILS: {_MATTERSIM_ERR}")
        if "pkg_resources" in _MATTERSIM_ERR:
            raise RuntimeError("MatterSim import failed: missing setuptools (pkg_resources). Please install with 'pip install setuptools'.")
        raise RuntimeError("MatterSim is not installed. Please reinstall with 'pip install macer'.")

    if model_path is None:
        default_mattersim_model_name = DEFAULT_MODELS.get("mattersim")
        if default_mattersim_model_name:
            # Check and Provision
            ensure_model("mattersim", default_mattersim_model_name)
            model_path = resolve_model_path(default_mattersim_model_name)
            print(f"No specific MatterSim model path provided; using default: {model_path}")
        else:
            raise ValueError("A model path (load_path) is required for MatterSim. No default model found in default-model.yaml.")
    else:
        # Check if the file exists. If not, try to auto-provision.
        if not os.path.exists(model_path):
            basename = os.path.basename(model_path)
            provisioned = ensure_model("mattersim", basename)
            if provisioned:
                model_path = provisioned
            else:
                # Provisioning failed or model unknown
                print_model_help("mattersim")
                raise FileNotFoundError(f"Model file not found: {model_path}")
        
        # If it was a relative path and exists (or was just provisioned), resolve it to absolute if needed
        # (Though ensure_model returns absolute path, and existing absolute path is fine)
        if not os.path.isabs(model_path):
             model_path = resolve_model_path(model_path)

    # MatterSimCalculator takes device and load_path arguments.
    return MatterSimCalculator(device=device, load_path=model_path)



"""Lattice reduction operations."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import numpy as np

try:
    from . import _spglib
except ImportError:
    try:
        from spglib import _spglib
    except ImportError:
        import _spglib

from .error import _set_no_error, _set_or_throw_error

__all__ = [
    "delaunay_reduce",
    "niggli_reduce",
]


def delaunay_reduce(
    lattice: ArrayLike[np.double], eps: float = 1e-5
) -> np.ndarray | None:
    r"""Run Delaunay reduction. When the search failed, `None` is returned.

    The transformation from original basis vectors
    :math:`( \mathbf{a} \; \mathbf{b} \; \mathbf{c} )`
    to final basis vectors :math:`( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )` is achieved by linear
    combination of basis vectors with integer coefficients without
    rotating coordinates. Therefore the transformation matrix is obtained
    by :math:`\mathbf{P} = ( \mathbf{a} \; \mathbf{b} \; \mathbf{c} ) ( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )^{-1}` and the matrix
    elements have to be almost integers.

    The algorithm is found in the international tables for crystallography volume A.

    Parameters
    ----------
    lattice: ndarray, (3, 3)
        Lattice parameters in the form of

        .. code-block::

            [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

    eps: float
        Tolerance parameter, but unlike `symprec` the unit is not a length.
        Tolerance to check if volume is close to zero or not and
        if two basis vectors are orthogonal by the value of dot
        product being close to zero or not.

    Returns
    -------
    delaunay_lattice: ndarray, (3, 3)
        Reduced lattice parameters are given as a numpy 'double' array:

        .. code-block::

            [[a_x, a_y, a_z],
             [b_x, b_y, b_z],
             [c_x, c_y, c_z]]

    Notes
    -----
    .. versionadded:: 1.9.4

    """  # noqa: E501
    _set_no_error()

    delaunay_lattice = np.array(np.transpose(lattice), dtype="double", order="C")
    try:
        _spglib.delaunay_reduce(delaunay_lattice, float(eps))
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return np.array(np.transpose(delaunay_lattice), dtype="double", order="C")


def niggli_reduce(
    lattice: ArrayLike[np.double], eps: float = 1e-5
) -> np.ndarray | None:
    r"""Run Niggli reduction. When the search failed, ``None`` is returned.

    The transformation from original basis vectors :math:`( \mathbf{a} \; \mathbf{b} \; \mathbf{c} )` to final basis vectors :math:`( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )` is achieved by linear
    combination of basis vectors with integer coefficients without
    rotating coordinates. Therefore the transformation matrix is obtained
    by :math:`\mathbf{P} = ( \mathbf{a} \; \mathbf{b} \; \mathbf{c} ) ( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )^{-1}` and the matrix
    elements have to be almost integers.

    The algorithm detail is found at https://atztogo.github.io/niggli/ and
    the references are there in.

    Parameters
    ----------
    lattice: ndarray
        Lattice parameters in the form of

        .. code-block::

            [[a_x, a_y, a_z],
            [b_x, b_y, b_z],
            [c_x, c_y, c_z]]

    eps: float
        Tolerance parameter, but unlike `symprec` the unit is not a length.
        This is used to check if difference of norms of two basis
        vectors is close to zero or not and if two basis vectors are
        orthogonal by the value of dot product being close to zero or
        not.
        The detail is shown at https://atztogo.github.io/niggli/.

    Returns
    -------
    niggli_lattice: ndarray, (3, 3)
        if the Niggli reduction succeeded:
            Reduced lattice parameters are given as a numpy 'double' array:

            .. code-block::

                [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

        otherwise None is returned.

    Notes
    -----
    .. versionadded:: 1.9.4

    """  # noqa: E501
    _set_no_error()

    niggli_lattice = np.array(np.transpose(lattice), dtype="double", order="C")
    try:
        _spglib.niggli_reduce(niggli_lattice, float(eps))
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return np.array(np.transpose(niggli_lattice), dtype="double", order="C")



import numpy as np
from scipy.optimize import curve_fit, minimize_scalar

h_planck = 4.135667662e-3  # eV/ps
h_planck_bar = 6.58211951e-4  # eV/ps
kb_boltzmann = 8.6173324e-5  # eV/K


def get_standard_errors_from_covariance(covariance):
   # return np.linalg.eigvals(covariance)
    return np.sqrt(np.diag(covariance))
    #return np.sqrt(np.trace(covariance))


class Lorentzian:
    def __init__(self,
                 test_frequencies_range,
                 power_spectrum,
                 guess_position=None,
                 guess_height=None):

        self.test_frequencies_range = test_frequencies_range
        self.power_spectrum = power_spectrum
        self.guess_pos = guess_position
        self.guess_height = guess_height

        self._fit_params = None
        self._fit_covariances = None

        self.curve_name = 'Lorentzian'

    def _function(self, x, a, b, c, d):
        """Lorentzian function
        x: frequency coordinate
        a: peak position
        b: half width
        c: area proportional parameter
        d: base line
        """
        return c/(np.pi*b*(1.0+((x - a)/b)**2))+d

    def get_fitting_parameters(self):

        if self._fit_params is None:
            if self.guess_pos is None or self.guess_height is None:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum)
            else:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum,
                                                        p0=[self.guess_pos, 0.1, self.guess_height, 0.0])
            self._fit_covariances = fit_covariances
            self._fit_params = fit_params

        return self._fit_params, self._fit_covariances

    def get_fitting(self):
        from scipy.integrate import quad

        try:

            fit_params, fit_covariances = self.get_fitting_parameters()

            maximum = fit_params[2]/(fit_params[1]*np.pi)
            width = 2.0*fit_params[1]
            frequency = fit_params[0]
            area = fit_params[2]

            standard_errors = get_standard_errors_from_covariance(fit_covariances)

            global_error = np.average(standard_errors[:2])/np.sqrt(area)
            if np.isnan(global_error):
                raise RuntimeError

            #error = get_error_from_covariance(fit_covariances)
            base_line = fit_params[3]


            return {'maximum': maximum,
                    'width': width,
                    'peak_position': frequency,
                    'standard_errors': standard_errors,
                    'global_error': global_error,
                    'area': area,
                    'base_line': base_line,
                    'all_good': True}

        except RuntimeError:
            return {'all_good': False}

    def get_curve(self, frequency_range):
        return self._function(frequency_range, *self.get_fitting_parameters()[0])


class Lorentzian_asymmetric:
    def __init__(self,
                 test_frequencies_range,
                 power_spectrum,
                 guess_position=None,
                 guess_height=None):

        self.test_frequencies_range = test_frequencies_range
        self.power_spectrum = power_spectrum
        self.guess_pos = guess_position
        self.guess_height = guess_height

        self._fit_params = None
        self._fit_covariances = None

        self.curve_name = 'Assym. Lorentzian'

    def _g_a (self, x, a, b, s):
        """Asymmetric width term
        x: frequency coordinate
        a: peak position
        b: half width
        s: asymmetry parameter
        """
        return 2*b/(1.0+np.exp(s*(x-a)))

    def _function(self, x, a, b, c, d, s):
        """Lorentzian asymmetric function
        x: frequency coordinate
        a: peak position
        b: half width
        c: area proportional parameter
        d: base line
        s: asymmetry parameter
        """
        return c/(np.pi*self._g_a(x, a, b, s)*(1.0+((x-a)/(self._g_a(x, a, b, s)))**2))+d

    def get_fitting_parameters(self):

        if self._fit_params is None:
            if self.guess_pos is None or self.guess_height is None:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum)
            else:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum,
                                                        p0=[self.guess_pos, 0.1, self.guess_height, 0.0, 0.0])
            self._fit_covariances = fit_covariances
            self._fit_params = fit_params

        return self._fit_params, self._fit_covariances


    def get_fitting(self):
        from scipy.integrate import quad

        try:

            fit_params, fit_covariances = self.get_fitting_parameters()

            peak_pos = minimize_scalar(lambda x: -self._function(x, *fit_params), fit_params[0],
                                       bounds=[self.test_frequencies_range[0], self.test_frequencies_range[-1]],
                                       method='bounded')

            frequency = peak_pos["x"]
            maximum = -peak_pos["fun"]
            width = 2.0 * self._g_a(frequency, fit_params[0], fit_params[1], fit_params[4])
            asymmetry = fit_params[4]

            area, error_integration = quad(self._function, 0, self.test_frequencies_range[-1],
                                           args=tuple(fit_params),
                                           epsabs=1e-8)
        #    area = fit_params[2]

            standard_errors = get_standard_errors_from_covariance(fit_covariances)
            global_error = np.average(standard_errors[:2])/np.sqrt(area)
            if np.isnan(global_error):
                raise RuntimeError

            #error = get_error_from_covariance(fit_covariances)
            base_line = fit_params[3]

            return {'maximum': maximum,
                    'width': width,
                    'peak_position': frequency,
                    'global_error': global_error,
                    'area': area,
                    'base_line': base_line,
                    'asymmetry': asymmetry,
                    'all_good': True}

        except RuntimeError:
            return {'all_good': False}

    def get_curve(self, frequency_range):
        return self._function(frequency_range, *self.get_fitting_parameters()[0])


class Damped_harmonic:
    def __init__(self,
                 test_frequencies_range,
                 power_spectrum,
                 guess_position=None,
                 guess_height=None):

        self.test_frequencies_range = test_frequencies_range
        self.power_spectrum = power_spectrum
        self.guess_pos = guess_position
        self.guess_height = guess_height

        self._fit_params = None
        self._fit_covariances = None

        self.curve_name = 'Damped Harm. Osc.'

    def _function(self, x, a, b, c, d):
        """Damped harmonic oscillator PS function
        x: frequency coordinate
        a: peak position
        b: half width
        c: area proportional parameter
        d: base line
        """
        return c/((a**2-x**2)**2 + (b*x)**2)+d

    def get_fitting_parameters(self):

        if self._fit_params is None:
            if self.guess_pos is None or self.guess_height is None:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum)
            else:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum,
                                                        p0=[self.guess_pos, 0.1, self.guess_height, 0.0])
            self._fit_covariances = fit_covariances
            self._fit_params = fit_params

        return self._fit_params, self._fit_covariances

    def get_fitting(self):
        from scipy.integrate import quad

        try:

            fit_params, fit_covariances = self.get_fitting_parameters()

            self._fit_params = fit_params

            width = abs(fit_params[1])

            maximum = fit_params[2]/(width*np.pi)
            frequency = fit_params[0]
            maximum = self.get_curve(frequency)


            area, error_integration = quad(self._function, 0, self.test_frequencies_range[-1],
                                           args=tuple(fit_params),
                                           epsabs=1e-8)
#            area = fit_params[2]*np.pi/(fit_params[0]**3*width)

            standard_errors = get_standard_errors_from_covariance(fit_covariances)

            global_error = np.average(standard_errors[:2])/np.sqrt(area)
            if np.isnan(global_error):
                raise RuntimeError

            base_line = fit_params[3]

            return {'maximum': maximum,
                    'width': width,
                    'peak_position': frequency,
                    'global_error': global_error,
                    'area': area,
                    'base_line': base_line,
                    'all_good': True}

        except RuntimeError:
            return {'all_good': False}

    def get_curve(self, frequency_range):
        return self._function(frequency_range, *self.get_fitting_parameters()[0])


class Gaussian_function:
    def __init__(self,
                 test_frequencies_range,
                 power_spectrum,
                 guess_position=None,
                 guess_height=None):

        self.test_frequencies_range = test_frequencies_range
        self.power_spectrum = power_spectrum
        self.guess_pos = guess_position
        self.guess_height = guess_height

        self._fit_params = None
        self._fit_covariances = None

        self.curve_name = 'Gaussian dist.'

    def _function(self, x, a, b, c, d):
        """Gaussian PDF function
        x: coordinate
        a: peak position
        b: deviation (sigma)
        c: area proportional parameter
        d: base line
        """
        return c/b*np.sqrt(2*np.pi)*np.exp(-(x-a)**2/(2*b**2))+d

    def get_fitting_parameters(self):

        if self._fit_params is None:
            if self.guess_pos is None or self.guess_height is None:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum)
            else:
                fit_params, fit_covariances = curve_fit(self._function,
                                                        self.test_frequencies_range,
                                                        self.power_spectrum,
                                                        p0=[self.guess_pos, 0.1, self.guess_height, 0.0])
            self._fit_covariances = fit_covariances
            self._fit_params = fit_params

        return self._fit_params, self._fit_covariances

    def get_fitting(self):
        from scipy.integrate import quad

        try:

            fit_params, fit_covariances = self.get_fitting_parameters()

            self._fit_params = fit_params

            width = abs(fit_params[1])

            frequency = fit_params[0]
            maximum = self.get_curve(frequency)

            area, error_integration = quad(self._function, 0, self.test_frequencies_range[-1],
                                           args=tuple(fit_params),
                                           epsabs=1e-8)
#            area = fit_params[2]*np.pi/(fit_params[0]**3*width)

            standard_errors = get_standard_errors_from_covariance(fit_covariances)

            global_error = np.average(standard_errors[:2])/np.sqrt(area)
            if np.isnan(global_error):
                raise RuntimeError

            base_line = fit_params[3]

            return {'maximum': maximum,
                    'width': width,
                    'peak_position': frequency,
                    'global_error': global_error,
                    'area': area,
                    'base_line': base_line,
                    'all_good': True}

        except RuntimeError:
            return {'all_good': False}

    def get_curve(self, frequency_range):
        return self._function(frequency_range, *self.get_fitting_parameters()[0])



fitting_functions = {
    0: Lorentzian,
    1: Lorentzian_asymmetric,
    2: Damped_harmonic,
}

# Test for automatic detection (order can change)
# import sys, inspect
# list_fitting = inspect.getmembers(sys.modules[__name__], inspect.isclass)
# Fitting_functions = {}
# for i, p in enumerate(list_fitting):
#    Fitting_functions[i] = p[1]



import argparse
import ast
import dataclasses
import logging
import os
import urllib.request
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

import torch

from macer.externals.mace_bundled.cli.fine_tuning_select import (
    FilteringType,
    SelectionSettings,
    SubselectType,
    select_samples,
)
from macer.externals.mace_bundled.data import AtomicData, KeySpecification
from macer.externals.mace_bundled.data.utils import Configuration
from macer.externals.mace_bundled.tools import torch_geometric
from macer.externals.mace_bundled.tools.scripts_utils import SubsetCollection, get_dataset_from_xyz
from macer.externals.mace_bundled.tools.utils import AtomicNumberTable, get_cache_dir


@dataclasses.dataclass
class HeadConfig:
    head_name: str
    key_specification: KeySpecification
    train_file: Optional[Union[str, List[str]]] = None
    valid_file: Optional[Union[str, List[str]]] = None
    test_file: Optional[str] = None
    test_dir: Optional[str] = None
    E0s: Optional[Any] = None
    statistics_file: Optional[str] = None
    valid_fraction: Optional[float] = None
    config_type_weights: Optional[Dict[str, float]] = None
    keep_isolated_atoms: Optional[bool] = None
    atomic_numbers: Optional[Union[List[int], List[str]]] = None
    mean: Optional[float] = None
    std: Optional[float] = None
    avg_num_neighbors: Optional[float] = None
    compute_avg_num_neighbors: Optional[bool] = None
    collections: Optional[SubsetCollection] = None
    train_loader: Optional[torch.utils.data.DataLoader] = None
    z_table: Optional[Any] = None
    atomic_energies_dict: Optional[Dict[str, float]] = None


def dict_head_to_dataclass(
    head: Dict[str, Any], head_name: str, args: argparse.Namespace
) -> HeadConfig:
    """Convert head dictionary to HeadConfig dataclass."""
    # parser+head args that have no defaults but are required
    if (args.train_file is None) and (head.get("train_file", None) is None):
        raise ValueError(
            "train file is not set in the head config yaml or via command line args"
        )

    return HeadConfig(
        head_name=head_name,
        train_file=head.get("train_file", args.train_file),
        valid_file=head.get("valid_file", args.valid_file),
        test_file=head.get("test_file", None),
        test_dir=head.get("test_dir", None),
        E0s=head.get("E0s", args.E0s),
        statistics_file=head.get("statistics_file", args.statistics_file),
        valid_fraction=head.get("valid_fraction", args.valid_fraction),
        config_type_weights=head.get("config_type_weights", args.config_type_weights),
        compute_avg_num_neighbors=head.get(
            "compute_avg_num_neighbors", args.compute_avg_num_neighbors
        ),
        atomic_numbers=head.get("atomic_numbers", args.atomic_numbers),
        mean=head.get("mean", args.mean),
        std=head.get("std", args.std),
        avg_num_neighbors=head.get("avg_num_neighbors", args.avg_num_neighbors),
        key_specification=head["key_specification"],
        keep_isolated_atoms=head.get("keep_isolated_atoms", args.keep_isolated_atoms),
    )


def prepare_default_head(args: argparse.Namespace) -> Dict[str, Any]:
    """Prepare a default head from args."""
    return {
        "Default": {
            "train_file": args.train_file,
            "valid_file": args.valid_file,
            "test_file": args.test_file,
            "test_dir": args.test_dir,
            "E0s": args.E0s,
            "statistics_file": args.statistics_file,
            "key_specification": args.key_specification,
            "valid_fraction": args.valid_fraction,
            "config_type_weights": args.config_type_weights,
            "keep_isolated_atoms": args.keep_isolated_atoms,
        }
    }


def prepare_pt_head(
    args: argparse.Namespace,
    pt_keyspec: KeySpecification,
    foundation_model_num_neighbours: float,
) -> Dict[str, Any]:
    """Prepare a pretraining head from args."""
    if (
        args.foundation_model in ["small", "medium", "large"]
        or args.pt_train_file == "mp"
    ):
        logging.info(
            "Using foundation model for multiheads finetuning with Materials Project data"
        )
        pt_keyspec.update(
            info_keys={"energy": "energy", "stress": "stress"},
            arrays_keys={"forces": "forces"},
        )
        pt_head = {
            "train_file": "mp",
            "E0s": "foundation",
            "statistics_file": None,
            "key_specification": pt_keyspec,
            "avg_num_neighbors": foundation_model_num_neighbours,
            "compute_avg_num_neighbors": False,
        }
    else:
        pt_head = {
            "train_file": args.pt_train_file,
            "valid_file": args.pt_valid_file,
            "E0s": "foundation",
            "statistics_file": args.statistics_file,
            "valid_fraction": args.valid_fraction,
            "key_specification": pt_keyspec,
            "avg_num_neighbors": foundation_model_num_neighbours,
            "keep_isolated_atoms": args.keep_isolated_atoms,
            "compute_avg_num_neighbors": False,
        }

    return pt_head


def assemble_replay_data(
    name: str,
    args: argparse.Namespace,
    head_config_pt: HeadConfig,
    tag: str,
) -> SubsetCollection:
    """Assemble data for replay fine-tuning."""
    try:
        if name == "mp":
            checkpoint_url = "https://github.com/ACEsuit/mace-foundations/releases/download/mace_mp_0b/mp_traj_combined.xyz"
        elif name == "matpes_pbe":
            checkpoint_url = "https://github.com/ACEsuit/mace-foundations/releases/download/mace_matpes_0/matpes-pbe-replay-data.xyz"
        elif name == "matpes_r2scan":
            checkpoint_url = "https://github.com/ACEsuit/mace-foundations/releases/download/mace_matpes_0/matpes-r2scan-replay-data.extxyz"
        else:
            raise ValueError(f"Unknown replay dataset name {name}")

        cache_dir = get_cache_dir()
        checkpoint_url_name = "".join(
            c for c in os.path.basename(checkpoint_url) if c.isalnum() or c in "_"
        )
        cached_dataset_path = f"{cache_dir}/{checkpoint_url_name}"
        if not os.path.isfile(cached_dataset_path):
            os.makedirs(cache_dir, exist_ok=True)
            # download and save to disk
            logging.info("Downloading MP structures for finetuning")
            _, http_msg = urllib.request.urlretrieve(
                checkpoint_url, cached_dataset_path
            )
            if "Content-Type: text/html" in http_msg:
                raise RuntimeError(
                    f"Dataset download failed, please check the URL {checkpoint_url}"
                )
            logging.info(f"Materials Project dataset to {cached_dataset_path}")
        output = f"mp_finetuning-{tag}.xyz"
        atomic_numbers = (
            ast.literal_eval(args.atomic_numbers)
            if args.atomic_numbers is not None
            else None
        )
        settings = SelectionSettings(
            configs_pt=cached_dataset_path,
            output=f"mp_finetuning-{tag}.xyz",
            atomic_numbers=atomic_numbers,
            num_samples=args.num_samples_pt,
            seed=args.seed,
            head_pt="pbe_mp",
            weight_pt=args.weight_pt_head,
            filtering_type=FilteringType(args.filter_type_pt),
            subselect=SubselectType(args.subselect_pt),
            default_dtype=args.default_dtype,
            allow_random_padding=args.allow_random_padding_pt,
        )
        select_samples(settings)
        head_config_pt.train_file = [output]
        collections_mp, _ = get_dataset_from_xyz(
            work_dir=args.work_dir,
            train_path=output,
            valid_path=None,
            valid_fraction=args.valid_fraction,
            config_type_weights=None,
            test_path=None,
            seed=args.seed,
            key_specification=head_config_pt.key_specification,
            head_name="pt_head",
            keep_isolated_atoms=args.keep_isolated_atoms,
            no_data_ok=(
                args.pseudolabel_replay
                and args.multiheads_finetuning
                and head_config_pt.head_name == "pt_head"
            ),
        )
        return collections_mp
    except Exception as exc:
        raise RuntimeError(
            "Foundation model replay data or descriptors cached data not found and download failed"
        ) from exc


def generate_pseudolabels_for_configs(
    model: torch.nn.Module,
    configs: List[Configuration],
    z_table: AtomicNumberTable,
    r_max: float,
    device: torch.device,
    batch_size: int,
) -> List[Configuration]:
    """
    Generate pseudolabels for a list of Configuration objects.

    Args:
        model: The foundation model
        configs: List of Configuration objects
        z_table: Atomic number table
        r_max: Cutoff radius
        device: Device to run model on
        batch_size: Batch size for inference

    Returns:
        List of Configuration objects with updated properties
    """

    model.eval()
    updated_configs = []

    # Disable gradient tracking for model parameters
    original_requires_grad = {}
    for param in model.parameters():
        original_requires_grad[param] = param.requires_grad
        param.requires_grad = False

    # Process configs in batches
    for i in range(0, len(configs), batch_size):
        batch_configs = configs[i : i + batch_size]

        try:
            # Create temporary AtomicData objects for this batch
            batch_data = [
                AtomicData.from_config(config, z_table=z_table, cutoff=r_max)
                for config in batch_configs
            ]

            # Create a batch for model inference
            batch = torch_geometric.Batch.from_data_list(batch_data).to(device)
            batch_dict = batch.to_dict()

            # Run model inference with computation of all properties
            out = model(
                batch_dict,
                training=False,
                compute_force=True,
                compute_virials=True,
                compute_stress=True,
            )

            # Process each configuration in the batch
            for j, config in enumerate(batch_configs):
                # Create a deepcopy to avoid modifying the original
                config_copy = deepcopy(config)

                # Ensure properties dict exists
                if not hasattr(config_copy, "properties"):
                    config_copy.properties = {}

                # Update config properties with pseudolabels
                if "energy" in out and out["energy"] is not None:
                    config_copy.properties["energy"] = (
                        out["energy"][j].detach().cpu().item()
                    )
                if "forces" in out and out["forces"] is not None:
                    # Forces are per atom
                    node_start = batch.ptr[j].item()
                    node_end = batch.ptr[j + 1].item()

                    config_copy.properties["forces"] = (
                        out["forces"][node_start:node_end].detach().cpu().numpy()
                    )
                if "stress" in out and out["stress"] is not None:
                    config_copy.properties["stress"] = (
                        out["stress"][j].detach().cpu().numpy()
                    )
                if "virials" in out and out["virials"] is not None:
                    config_copy.properties["virials"] = (
                        out["virials"][j].detach().cpu().numpy()
                    )
                if "dipole" in out and out["dipole"] is not None:
                    config_copy.properties["dipole"] = (
                        out["dipole"][j].detach().cpu().numpy()
                    )
                if "charges" in out and out["charges"] is not None:
                    # Charges are per atom
                    node_start = batch.ptr[j].item()
                    node_end = batch.ptr[j + 1].item()

                    config_copy.properties["charges"] = (
                        out["charges"][node_start:node_end].detach().cpu().numpy()
                    )

                updated_configs.append(config_copy)

        except Exception as e:  # pylint: disable=broad-except
            logging.error(
                f"Error generating pseudolabels for batch {i//batch_size + 1}: {str(e)}"
            )
            # On error, return the original configs for this batch
            updated_configs.extend([deepcopy(config) for config in batch_configs])

    # Restore original requires_grad settings
    for param, requires_grad in original_requires_grad.items():
        param.requires_grad = requires_grad

    logging.info(f"Generated pseudolabels for {len(updated_configs)} configurations")
    return updated_configs


def apply_pseudolabels_to_pt_head_configs(
    foundation_model: torch.nn.Module,
    pt_head_config: HeadConfig,
    r_max: float,
    device: torch.device,
    batch_size: int,
) -> bool:
    """
    Apply pseudolabels to pt_head configurations using the foundation model.

    Args:
        foundation_model: The pre-loaded foundation model
        pt_head_config: The HeadConfig object for pt_head
        r_max: Cutoff radius
        device: Device to run model on
        batch_size: Batch size for inference

    Returns:
        bool: True if pseudolabeling was successful, False otherwise
    """

    try:
        logging.info(
            "Applying pseudolabels to pt_head configurations using foundation model"
        )

        foundation_model.to(device)

        # Use foundation model's z_table if available
        if hasattr(foundation_model, "atomic_numbers"):
            z_table = AtomicNumberTable(
                sorted(foundation_model.atomic_numbers.tolist())
            )
            logging.info(
                f"Using foundation model's atomic numbers for pseudolabeling: {z_table.zs}"
            )
        elif hasattr(pt_head_config, "z_table") and pt_head_config.z_table is not None:
            z_table = pt_head_config.z_table
            logging.info(f"Using pt_head's z_table for pseudolabeling: {z_table.zs}")
        else:
            logging.warning("No atomic number table available for pseudolabeling")
            return False

        # Process training configurations
        if (
            hasattr(pt_head_config.collections, "train")
            and pt_head_config.collections.train
        ):
            logging.info(
                f"Generating pseudolabels for {len(pt_head_config.collections.train)} pt_head training configurations"
            )
            updated_train_configs = generate_pseudolabels_for_configs(
                model=foundation_model,
                configs=pt_head_config.collections.train,
                z_table=z_table,
                r_max=r_max,
                device=device,
                batch_size=batch_size,
            )

            # Replace the original configurations with updated ones
            pt_head_config.collections.train = updated_train_configs
            logging.info(
                f"Successfully applied pseudolabels to {len(updated_train_configs)} training configurations"
            )

        # Process validation configurations if they exist
        if (
            hasattr(pt_head_config.collections, "valid")
            and pt_head_config.collections.valid
        ):
            logging.info(
                f"Generating pseudolabels for {len(pt_head_config.collections.valid)} pt_head validation configurations"
            )
            updated_valid_configs = generate_pseudolabels_for_configs(
                model=foundation_model,
                configs=pt_head_config.collections.valid,
                z_table=z_table,
                r_max=r_max,
                device=device,
                batch_size=batch_size,
            )

            # Replace the original configurations with updated ones
            pt_head_config.collections.valid = updated_valid_configs
            logging.info(
                f"Successfully applied pseudolabels to {len(updated_valid_configs)} validation configurations"
            )

        return True

    except Exception as e:  # pylint: disable=broad-except
        logging.error(f"Error applying pseudolabels: {str(e)}")
        return False



"""KPoints related operations."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike

try:
    from . import _spglib
except ImportError:
    try:
        from spglib import _spglib
    except ImportError:
        import _spglib

from .error import _set_no_error, _set_or_throw_error
from .utils import Cell, _expand_cell

__all__ = [
    "get_ir_reciprocal_mesh",
]


def get_ir_reciprocal_mesh(
    mesh: ArrayLike[np.intc],
    cell: Cell,
    is_shift: ArrayLike[np.intc] | None = None,
    is_time_reversal: bool = True,
    symprec: float = 1e-5,
    is_dense: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """Return k-points mesh and k-point map to the irreducible k-points.

    The symmetry is searched from the input cell.

    Parameters
    ----------
    mesh : array_like
        Uniform sampling mesh numbers.
        dtype='intc', shape=(3,)
    cell : spglib cell tuple
        Crystal structure.
    is_shift : array_like, optional
        [0, 0, 0] gives Gamma center mesh and value 1 gives half mesh shift.
        Default is None which equals to [0, 0, 0].
        dtype='intc', shape=(3,)
    is_time_reversal : bool, optional
        Whether time reversal symmetry is included or not. Default is True.
    symprec : float, optional
        Symmetry tolerance in distance. Default is 1e-5.
    is_dense : bool, optional
        grid_mapping_table is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    grid_mapping_table : ndarray
        Grid point mapping table to ir-gird-points.
        dtype='intc' or 'uintp', shape=(prod(mesh),)
    grid_address : ndarray
        Address of all grid points.
        dtype='intc', shape=(prod(mesh), 3)

    """
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)
    if lattice is None:
        return None

    if is_dense:
        dtype = "uintp"
    else:
        dtype = "intc"
    grid_mapping_table = np.zeros(np.prod(mesh), dtype=dtype)
    grid_address = np.zeros((np.prod(mesh), 3), dtype="intc")
    if is_shift is None:
        is_shift = [0, 0, 0]
    try:
        _spglib.ir_reciprocal_mesh(
            grid_address,
            grid_mapping_table,
            np.array(mesh, dtype="intc"),
            np.array(is_shift, dtype="intc"),
            int(is_time_reversal * 1),
            lattice,
            positions,
            numbers,
            float(symprec),
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return grid_mapping_table, grid_address



"""Symmetry adapted basis sets of 3rd order force constants."""

from __future__ import annotations

import time
from typing import Optional, Union

import numpy as np
from scipy.sparse import coo_array, csr_array

from macer.externals.symfc_bundled.spg_reps import SpgRepsO3
from macer.externals.symfc_bundled.utils.eig_tools import (
    eigsh_projector,
    eigsh_projector_sumrule,
)
from macer.externals.symfc_bundled.utils.matrix import BlockMatrixNode

try:
    from macer.externals.symfc_bundled.utils.matrix import dot_product_sparse
except ImportError:
    pass

from macer.externals.symfc_bundled.utils.permutation_tools_O3 import compr_permutation_lat_trans_O3
from macer.externals.symfc_bundled.utils.translation_tools_O3 import compressed_projector_sum_rules_O3
from macer.externals.symfc_bundled.utils.utils import SymfcAtoms
from macer.externals.symfc_bundled.utils.utils_O3 import (
    get_atomic_lat_trans_decompr_indices_O3,
    get_compr_coset_projector_O3,
    get_lat_trans_compr_matrix_O3,
)

from . import FCBasisSetBase


def print_sp_matrix_size(c: Union[csr_array, coo_array], header: str):
    """Show sparse matrix size."""
    print(header, c.shape, len(c.data), flush=True)


class FCBasisSetO3(FCBasisSetBase):
    r"""Symmetry adapted basis set for 3rd order force constants.

    Attributes
    ----------
    basis_set : ndarray
        Compressed force constants basis set. The first dimension n_compr
        (<< 27 * N ** 3, \sim n_bases) is given as a result of compression,
        which depends on the system.  shape=(n_compr, n_bases), dtype='double'
    n_a_compression_matrix : csr_array
        Compression matrix compressed by lattice translation. The basis set
        compressed only by lattice translation is obtained by
        n_a_compression_matrix @ basis_set.
        shape=(n_a * N * N * 27, n_compr), dtype='double'
    translation_permutations : ndarray
        Atom indices after lattice translations.
        shape=(lattice_translations, supercell_atoms), dtype=int.

    """

    def __init__(
        self,
        supercell: SymfcAtoms,
        cutoff: Optional[float] = None,
        spacegroup_operations: Optional[dict] = None,
        use_mkl: bool = False,
        log_level: int = 0,
    ):
        """Init method.

        Parameters
        ----------
        supercell : SymfcAtoms
            Supercell.
        cutoff: float
            Cutoff distance in angstroms. Default is None.
        spacegroup_operations : dict, optional
            Space group operations in supercell, by default None. When None,
            spglib is used. The following keys and values correspond to spglib
            symmetry dataset:
                rotations : array_like
                translations : array_like
        use_mkl : bool
            Use MKL or not. Default is False.
        log_level : int, optional
            Log level. Default is 0.

        """
        super().__init__(supercell, cutoff=cutoff, use_mkl=use_mkl, log_level=log_level)
        self._spg_reps = SpgRepsO3(
            supercell, spacegroup_operations=spacegroup_operations
        )

        trans_perms = self._spg_reps.translation_permutations
        self._atomic_decompr_idx = get_atomic_lat_trans_decompr_indices_O3(trans_perms)

        self._n_a_compression_matrix: Optional[csr_array] = None
        self._basis_set: Optional[np.ndarray] = None
        self._blocked_basis_set: Optional[BlockMatrixNode] = None

    @property
    def compression_matrix(self) -> Optional[csr_array]:
        """Return compression matrix.

        This expands fc basis_sets to (N*N*N*3*3*3, n_bases).

        """
        if self._n_a_compression_matrix is None:
            raise ValueError(
                "Compression matrix is not computed yet. "
                "Call run() method to compute it."
            )
        trans_perms = self._spg_reps.translation_permutations
        c_trans = get_lat_trans_compr_matrix_O3(trans_perms)
        return dot_product_sparse(
            c_trans, self._n_a_compression_matrix, use_mkl=self._use_mkl
        )

    @property
    def compact_compression_matrix(self) -> Optional[csr_array]:
        """Return compact compression matrix.

        This expands fc basis_sets to (n_a*N*N*3*3*3, n_bases).

        """
        if self._n_a_compression_matrix is None:
            raise ValueError(
                "Compression matrix is not computed yet. "
                "Call run() method to compute it."
            )
        n_lp = self.translation_permutations.shape[0]
        return self._n_a_compression_matrix / np.sqrt(n_lp)

    def run(self) -> FCBasisSetO3:
        """Compute compressed force constants basis set."""
        trans_perms = self._spg_reps.translation_permutations

        tt0 = time.time()
        c_pt = compr_permutation_lat_trans_O3(
            trans_perms,
            atomic_decompr_idx=self._atomic_decompr_idx,
            fc_cutoff=self._fc_cutoff,
            verbose=self._log_level > 0,
        )

        if self._log_level:
            print(" c_pt (size) :", c_pt.shape, flush=True)
        tt2 = time.time()

        proj_rpt = get_compr_coset_projector_O3(
            self._spg_reps,  # type: ignore
            fc_cutoff=self._fc_cutoff,
            atomic_decompr_idx=self._atomic_decompr_idx,
            c_pt=c_pt,
            use_mkl=self._use_mkl,
            verbose=self._log_level > 0,
        )
        tt3 = time.time()

        c_rpt = eigsh_projector(proj_rpt, verbose=self._log_level > 0)
        if self._log_level:
            print(" c_rpt (size) :", c_rpt.shape, flush=True)
        tt4 = time.time()

        n_a_compress_mat = dot_product_sparse(c_pt, c_rpt, use_mkl=self._use_mkl)
        tt5 = time.time()

        proj = compressed_projector_sum_rules_O3(
            trans_perms,
            n_a_compress_mat,
            atomic_decompr_idx=self._atomic_decompr_idx,
            fc_cutoff=self._fc_cutoff,
            use_mkl=self._use_mkl,
            verbose=self._log_level > 0,
        )
        tt6 = time.time()
        eigvecs = eigsh_projector_sumrule(
            proj,
            verbose=self._log_level > 0,
            use_mkl=self._use_mkl,
        )

        if self._log_level:
            print("Final size of basis set:", eigvecs.shape, flush=True)
        tt7 = time.time()

        if self._log_level:
            print(
                "Time (perm @ ltrans)               :",
                "{:.3f}".format(tt2 - tt0),
                flush=True,
            )
            print(
                "Time (coset)                       :",
                "{:.3f}".format(tt3 - tt2),
                flush=True,
            )
            print(
                "Time (eigh(coset @ perm @ ltrans)) :",
                "{:.3f}".format(tt4 - tt3),
                flush=True,
            )
            print(
                "Time (c_pt @ c_rpt)                :",
                "{:.3f}".format(tt5 - tt4),
                flush=True,
            )
            print(
                "Time (proj(sum))                   :",
                "{:.3f}".format(tt6 - tt5),
                flush=True,
            )
            print(
                "Time (eigh(sum))                   :",
                "{:.3f}".format(tt7 - tt6),
                flush=True,
            )
            print("---", flush=True)
            print(
                "Time (Basis FC3)                   :",
                "{:.3f}".format(tt7 - tt0),
                flush=True,
            )

        self._blocked_basis_set = eigvecs
        self._n_a_compression_matrix = n_a_compress_mat

        return self

    def estimate_basis_size(self) -> int:
        """Estimate basis set size."""
        if self._fc_cutoff is None:
            n_sym, N = self._spg_reps._permutations.shape
            basis_size_estimates = 27 * (N**3) / n_sym / 6
            return int(np.round(basis_size_estimates).astype(int))

        trans_perms = self._spg_reps.translation_permutations
        c_pt = compr_permutation_lat_trans_O3(
            trans_perms,
            atomic_decompr_idx=self._atomic_decompr_idx,
            fc_cutoff=self._fc_cutoff,
            verbose=False,
        )
        n_sym_prim = len(self._spg_reps._unique_rotations)
        basis_size_estimates = c_pt.shape[1] / n_sym_prim  # type: ignore
        return int(np.round(basis_size_estimates).astype(int))



#!/usr/bin/env python3
import h5py
import numpy as np
import argparse
import sys
import os

def main():
    parser = argparse.ArgumentParser(description="Extract Gruneisen parameters from HDF5 to text format.")
    parser.add_argument("input", nargs='?', default="gruneisen.hdf5", help="Input HDF5 file (default: gruneisen.hdf5)")
    parser.add_argument("-o", "--output", default="gruneisen_full.dat", help="Output data file (default: gruneisen_full.dat)")
    
    args = parser.parse_args()

    input_file = args.input
    output_file = args.output

    if not os.path.isfile(input_file):
        print(f"Error: Input file '{input_file}' not found.")
        sys.exit(1)

    print(f"Reading from {input_file}...")

    try:
        with h5py.File(input_file, 'r') as f:
            # Check required keys
            required_keys = ['qpoint', 'weight', 'frequency', 'gruneisen']
            for key in required_keys:
                if key not in f:
                    print(f"Error: Key '{key}' not found in HDF5 file.")
                    sys.exit(1)

            # Read data
            qpoints = f['qpoint'][:]     # (N_q, 3)
            weights = f['weight'][:]     # (N_q,)
            frequencies = f['frequency'][:] # (N_q, N_band)
            gruneisen = f['gruneisen'][:]   # (N_q, N_band)

            n_q, n_band = frequencies.shape
            
            print(f"Found {n_q} q-points and {n_band} bands.")
            print(f"Writing to {output_file}...")

            with open(output_file, 'w') as out:
                out.write("# qx       qy       qz       weight band_index frequency(THz)  gruneisen_param\n")
                for i in range(n_q):
                    qx, qy, qz = qpoints[i]
                    w = weights[i]
                    for j in range(n_band):
                        freq = frequencies[i, j]
                        gamma = gruneisen[i, j]
                        out.write(f"{qx:10.6f} {qy:10.6f} {qz:10.6f} "
                                  f"{w:6d} {j+1:4d} {freq:15.8f} {gamma:15.8f}\n")
            
            print("Done.")

    except Exception as e:
        print(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()



###########################################################################################
# Slurm environment setup for distributed training.
# This code is refactored from rsarm's contribution at:
# https://github.com/Lumi-supercomputer/lumi-reframe-tests/blob/main/checks/apps/deeplearning/pytorch/src/pt_distr_env.py
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import os

try:
    import hostlist
except ImportError:
    hostlist = None  # Only needed on SLURM systems


class DistributedEnvironment:
    def __init__(self):
        self._setup_distr_env()
        self.master_addr = os.environ["MASTER_ADDR"]
        self.master_port = os.environ["MASTER_PORT"]
        self.world_size = int(os.environ["WORLD_SIZE"])
        self.local_rank = int(os.environ["LOCAL_RANK"])
        self.rank = int(os.environ["RANK"])

    def _setup_distr_env(self):
        if "SLURM_JOB_NODELIST" in os.environ:
            hostname = hostlist.expand_hostlist(os.environ["SLURM_JOB_NODELIST"])[0]
            os.environ["MASTER_ADDR"] = hostname
            os.environ["MASTER_PORT"] = os.environ.get("MASTER_PORT", "33333")
            os.environ["WORLD_SIZE"] = os.environ.get(
                "SLURM_NTASKS",
                str(
                    int(os.environ["SLURM_NTASKS_PER_NODE"])
                    * int(os.environ["SLURM_NNODES"])
                ),
            )
            os.environ["LOCAL_RANK"] = os.environ["SLURM_LOCALID"]
            os.environ["RANK"] = os.environ["SLURM_PROCID"]
        else:
            # Assume local manual run with torchrun
            os.environ.setdefault("MASTER_ADDR", "localhost")
            os.environ.setdefault("MASTER_PORT", "33333")
            os.environ.setdefault("WORLD_SIZE", "1")
            os.environ.setdefault("LOCAL_RANK", "0")
            os.environ.setdefault("RANK", "0")

    def __repr__(self):
        return (
            f"DistributedEnvironment(master_addr={self.master_addr}, master_port={self.master_port}, "
            f"world_size={self.world_size}, local_rank={self.local_rank}, rank={self.rank})"
        )
